import os
import cv2
import json
import numpy as np
import pandas as pd
import scanpy as sc

import torch
import torch.nn as nn
import torch.optim as optim

from scipy.stats import median_abs_deviation
from torch.utils.data import DataLoader

import sys



# Module import
from ._starfysh import LOGGER
from .dataloader import VisiumDataset, VisiumPoEDataSet
from ._starfysh import AVAE, AVAE_PoE, train, train_poe


# -------------------
# Model Parameters
# -------------------

class VisiumArguments:
    """
    Loading Visium AnnData, perform preprocessing, library-size smoothing & Anchor spot detection

    Parameters
    ----------
    adata : AnnData
        annotated visium count matrix

    adata_norm : AnnData
        annotated visium count matrix after normalization & log-transform

    gene_sig : pd.DataFrame
        list of signature genes for each cell type. (dim: [S, Cell_type])

    img_metadata : dict
        Spatial information metadata (histology image, coordinates, scalefactor)
    """
    def __init__(
        self,
        adata,
        adata_norm,
        gene_sig,
        img_metadata,
        **kwargs
    ):

        self.adata = adata
        self.adata_norm = adata_norm
        self.gene_sig = gene_sig
        self.map_info = img_metadata['map_info'].iloc[:, :4].astype(float)
        self.scalefactor = img_metadata['scalefactor']
        self.img = img_metadata['img']
        self.img_patches = None 
        self.eps = 1e-6

        self.params = {
            'sample_id': 'ST', 
            'n_anchors': int(adata.shape[0]),
            'patch_r': 16,
            'signif_level': 3,
            'window_size': 1,
            'n_img_chan': 1
        }

        # Update parameters for library smoothing & anchor spot identification
        for k, v in kwargs.items():
            if k in self.params.keys():
                self.params[k] = v

        # Center expression for gene score calculation
        adata_scale = self.adata_norm.copy()
        sc.pp.scale(adata_scale)
        
        # Store cell types
        self.adata.uns['cell_types'] = list(self.gene_sig.columns)

        # Filter out signature genes X listed in expression matrix
        LOGGER.info('Subsetting highly variable & signature genes ...')
        self.adata, self.adata_norm = get_adata_wsig(adata, adata_norm, gene_sig)
        self.adata_scale = adata_scale[:, adata.var_names]

        # Update spatial metadata
        self._update_spatial_info(self.params['sample_id'])

        # Get smoothed library size
        LOGGER.info('Smoothing library size by taking averaging with neighbor spots...')
        log_lib = np.log1p(self.adata.X.sum(1))
        self.log_lib = np.squeeze(np.asarray(log_lib)) if log_lib.ndim > 1 else log_lib
        
        self.win_loglib = get_windowed_library(self.adata,
                                               self.map_info,
                                               self.log_lib,
                                               window_size=self.params['window_size'])

        # Retrieve & normalize signature gene expressions by 
        # comparing w/ randomized base expression (`sc.tl.score_genes`) 
        LOGGER.info('Retrieving & normalizing signature gene expressions...')
        self.sig_mean = self._get_sig_mean()
        self.sig_mean_norm = self._calc_gene_scores()
       
        # Get anchor spots
        LOGGER.info('Identifying anchor spots (highly expression of specific cell-type signatures)...')
        anchor_info = self._compute_anchors()

        # row-norm; "ReLU" on signature scores for valid dirichlet param
        self.sig_mean_norm[self.sig_mean_norm < 0] = self.eps
        self.sig_mean_norm = self.sig_mean_norm.div(self.sig_mean_norm.sum(1), axis=0)
        self.sig_mean_norm.fillna(1/self.sig_mean_norm.shape[1], inplace=True)

        self.pure_spots, self.pure_dict, self.pure_idx = anchor_info        
        del self.adata.raw, self.adata_norm.raw 

    def get_adata(self):
        """Return adata after preprocessing & HVG gene selection"""
        return self.adata, self.adata_norm

    def get_anchors(self):
        """Return indices of anchor spots for each cell type"""
        anchors_df = pd.DataFrame.from_dict(self.pure_dict, orient='index')
        anchors_df = anchors_df.transpose()
        
        # Check whether empty anchors detected for any factor
        empty_indices = np.where(
            (~pd.isna(anchors_df)).sum(0) == 0
        )[0]
        
        if len(empty_indices) > 0:
            raise ValueError("Cell type(s) {} has no anchors significantly enriched for its signatures,"
                             "please lower outlier stats `signif_level` or try zscore-based signature".format(
                                 anchors_df.columns[empty_indices].to_list()
                            ))
        
        return anchors_df.applymap(
            lambda x:
            -1 if x is None else np.where(self.adata.obs.index == x)[0][0]
        )

    def get_img_patches(self):
        assert self.img_patches is not None, "Please run Starfysh PoE first"
        return self.img_patches

    def append_factors(self, arche_markers):
        """
        Append list of archetypes (w/ corresponding markers) as additional cell type(s) / state(s) to the `gene_sig`
        """
        self.gene_sig = pd.concat((self.gene_sig, arche_markers), axis=1)

        # Update factor names & anchor spots
        self.adata.uns['cell_types'] = list(self.gene_sig.columns)
        self._update_anchors()
        return None

    def replace_factors(self, factors_to_repl, arche_markers):
        """
        Replace factor(s) with archetypes & their corresponding markers in the `gene_sig`
        """
        if isinstance(factors_to_repl, str):
            assert isinstance(arche_markers, pd.Series),\
                "Please pick only one archetype to replace the factor {}".format(factors_to_repl)
            factors_to_repl = [factors_to_repl]
            archetypes = [arche_markers.name]
        else:
            assert len(factors_to_repl) == len(arche_markers.columns), \
                "Unequal # cell types & archetypes to replace with"
            archetypes = arche_markers.columns

        self.gene_sig.rename(
            columns={
                f: a
                for (f, a) in zip(factors_to_repl, archetypes)
            }, inplace=True
        )
        self.gene_sig[archetypes] = pd.DataFrame(arche_markers)

        # Update factor names & anchor spots
        self.adata.uns['cell_types'] = list(self.gene_sig.columns)
        self._update_anchors()
        return None

    # --- Private methods ---
    def _compute_anchors(self):
        """
        Calculate top `anchor_spots` significantly enriched for given cell type(s)
        determined by gene set scores from signatures
        """
        score_df = self.sig_mean_norm
        n_anchor = self.params['n_anchors']

        top_expr_spots = (-score_df.values).argsort(axis=0)[:n_anchor, :]
        pure_spots = np.transpose(np.array(score_df.index)[top_expr_spots])

        pure_dict = {
            ct: spot
            for (spot, ct) in zip(pure_spots, score_df.columns)
        }

        pure_indices = np.zeros([score_df.shape[0], 1])
        idx = [np.where(score_df.index == i)[0][0] 
               for i in sorted({x for v in pure_dict.values() for x in v})]
        pure_indices[idx] = 1
        return pure_spots, pure_dict, pure_indices   
    
    def _update_anchors(self):
        """Re-calculate anchor spots given updated gene signatures"""
        self.sig_mean = self._get_sig_mean()
        self.sig_mean_norm = self._calc_gene_scores()
        self.adata.uns['cell_types'] = list(self.gene_sig.columns)

        LOGGER.info('Recalculating anchor spots (highly expression of specific cell-type signatures)...')
        anchor_info = self._compute_anchors()
        self.sig_mean_norm[self.sig_mean_norm < 0] = self.eps
        self.sig_mean_norm.fillna(1/self.sig_mean_norm.shape[1], inplace=True)
        self.pure_spots, self.pure_dict, self.pure_idx = anchor_info
              
    def _get_sig_mean(self):
        sig_mean_expr = pd.DataFrame()
        cnt_df = self.adata_norm.to_df()
                    
        # Calculate avg. signature expressions for each cell type
        for i, cell_type in enumerate(self.gene_sig.columns):
            sigs = np.intersect1d(cnt_df.columns, self.gene_sig.iloc[:, i].astype(str))
            
            if len(sigs) == 0:
                raise ValueError("Empty signatures for {},"
                                 "please double check your `gene_sig` input or set a higher"
                                 "`n_gene` threshold upon dataloading".format(cell_type))
                
            else:
                sig_mean_expr[cell_type] = cnt_df.loc[:, sigs].mean(axis=1)
        
        sig_mean_expr.index = self.adata.obs_names
        sig_mean_expr.columns = self.gene_sig.columns
        return sig_mean_expr

    def _update_spatial_info(self, sample_id):
        """Update paired spatial information to ST adata"""
        # Update image channel count for RGB input (`y`)
        if self.img is not None and self.img.ndim == 3: 
            self.params['n_img_chan'] = 3

        if 'spatial' not in self.adata.uns_keys():
            self.adata.uns['spatial'] = {
                sample_id: {
                    'images': {'hires': (self.img - self.img.min()) / (self.img.max() - self.img.min())},
                    'scalefactors': self.scalefactor
                },
            }

            self.adata_norm.uns['spatial'] = {
                sample_id: {
                    'images': {'hires': (self.img - self.img.min()) / (self.img.max() - self.img.min())},
                    'scalefactors': self.scalefactor
                },
            }

            self.adata.obsm['spatial'] = self.map_info[['imagecol', 'imagerow']].values
            self.adata_norm.obsm['spatial'] = self.map_info[['imagecol', 'imagerow']].values

        # Typecast: spatial coords.
        self.adata_norm.obsm['spatial'] = self.map_info[['imagecol', 'imagerow']].values
        self.adata_norm.obsm['spatial'] = self.adata_norm.obsm['spatial'].astype(np.float32) 
        return None

    def _update_img_patches(self, dl_poe):
        imgs = torch.Tensor(dl_poe.spot_img_stack)
        self.img_patches = imgs.reshape(imgs.shape[0], -1)
        return None

    def _norm_sig(self):
        # col-norm for each cell type: divided by mean
        gexp = self.sig_mean.apply(lambda x: x / x.mean(), axis=0)
        return gexp
        
    def _calc_gene_scores(self):
        """Calculate gene set enrichment scores for each signature sets"""
        adata = self.adata_scale.copy()
        #adata = self.adata_norm.copy()
        for cell_type in self.gene_sig.columns:
            sig = self.gene_sig[cell_type][~pd.isna(self.gene_sig[cell_type])].to_list()
            sc.tl.score_genes(adata, sig, use_raw=False, score_name=cell_type+'_score')
            
        gsea_df = adata.obs[[cell_type+'_score' for cell_type in self.gene_sig.columns]]
        gsea_df.columns = self.gene_sig.columns
        return gsea_df

# --------------------------------
# Running starfysh with 3-restart
# --------------------------------

def init_weights(module):
    if type(module) == nn.Linear:
        nn.init.xavier_uniform_(module.weight)

    elif type(module) == nn.BatchNorm1d:
        module.bias.data.zero_()
        module.weight.data.fill_(1.0)


def run_starfysh(
    visium_args,
    n_repeats=3,
    lr=1e-4,
    epochs=100,
    batch_size=32,
    alpha_mul=50,
    poe=False,
    device=torch.device('cpu'),
    seed=0,
    verbose=True
):
    """
    Wrapper to run starfysh deconvolution.
    
    Parameters
    ----------
    visium_args : VisiumArguments
        Preprocessed metadata calculated from input visium matrix:
        e.g. mean signature expression, library size, anchor spots, etc.

    n_repeats : int
        Number of restart to run Starfysh

    epochs : int
        Max. number of iterations

    poe : bool
        Whether to perform inference with Poe w/ image integration

    Returns
    -------
    best_model : starfysh.AVAE or starfysh.AVAE_PoE
        Trained Starfysh model with deconvolution results

    loss : np.ndarray
        Training losses
    """
    np.random.seed(seed)

    # Loading parameters
    adata = visium_args.adata
    win_loglib = visium_args.win_loglib
    gene_sig, sig_mean_norm = visium_args.gene_sig, visium_args.sig_mean_norm

    models = [None] * n_repeats
    losses = []
    loss_c_list = np.repeat(np.inf, n_repeats)

    if poe:
        dl_func = VisiumPoEDataSet  # dataloader
        train_func = train_poe  # training wrapper
    else:
        dl_func = VisiumDataset
        train_func = train

    trainset = dl_func(adata=adata, args=visium_args)
    trainloader = DataLoader(trainset, batch_size=batch_size, shuffle=True, drop_last=True)

    # Running Starfysh with multiple starts
    LOGGER.info('Running Starfysh with {} restarts, choose the model with best parameters...'.format(n_repeats))
    
    count = 0
    while count < n_repeats:
        best_loss_c = np.inf
        if poe:
            model = AVAE_PoE(
                adata=adata,
                gene_sig=sig_mean_norm,
                patch_r=visium_args.params['patch_r'],
                win_loglib=win_loglib,
                alpha_mul=alpha_mul,
                n_img_chan=visium_args.params['n_img_chan']
            )
            # Update patched & flattened image patches
            visium_args._update_img_patches(trainset)
        else:
            model = AVAE(
                adata=adata,
                gene_sig=sig_mean_norm,
                win_loglib=win_loglib,
                alpha_mul=alpha_mul
            )

        model = model.to(device)
        loss_dict = {
            'reconst': [],
            'c': [],
            'u': [],
            'z': [],
            'n': [],
            'tot': []
        }

        # Initialize model params
        if verbose:
            LOGGER.info('Initializing model parameters...')
            
        model.apply(init_weights)
        optimizer = optim.Adam(model.parameters(), lr=lr)
        scheduler = optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.98)
        
        try:
            from tqdm import tqdm

            # Initialize the tqdm progress bar
            progress_bar = tqdm(range(epochs), desc="Training Epochs")

            for epoch in progress_bar:
                result = train_func(model, trainloader, device, optimizer)
                torch.cuda.empty_cache()

                loss_tot, loss_reconst, loss_u, loss_z, loss_c, loss_n, corr_list = result
                if loss_c < best_loss_c:
                    models[count] = model
                    best_loss_c = loss_c

                torch.cuda.empty_cache()

                loss_dict['tot'].append(loss_tot)
                loss_dict['reconst'].append(loss_reconst)
                loss_dict['u'].append(loss_u)
                loss_dict['z'].append(loss_z)
                loss_dict['c'].append(loss_c)
                loss_dict['n'].append(loss_n)

                # Log progress every 10 epochs
                if (epoch + 1) % 10 == 0 and verbose:
                    log_message = "Epoch[{}/{}], train_loss: {:.4f}, train_reconst: {:.4f}, train_u: {:.4f},train_z: {:.4f},train_c: {:.4f},train_l: {:.4f}".format(
                        epoch + 1, epochs, loss_tot, loss_reconst, loss_u, loss_z, loss_c, loss_n)
                    progress_bar.write(log_message)

                scheduler.step()

            losses.append(loss_dict)
            loss_c_list[count] = best_loss_c

            count += 1
        except ValueError as ve: # Bad model initialization -> numerical instability
            continue
        
        if verbose:
                LOGGER.info('Saving the best-performance model...')
                LOGGER.info(" === Finished training === \n")

    idx = np.argmin(loss_c_list)
    best_model = models[idx]
    loss = losses[idx]

    return best_model, loss


# -------------------
# Preprocessing & IO
# -------------------

def preprocess(
    adata_raw,
    min_perc=None,
    max_perc=None,
    n_top_genes=2000,
    mt_thld=100,
    verbose=True,
    multiple_data=False
):
    """
    Preprocessing ST gexp matrix, remove Ribosomal & Mitochondrial genes

    Parameters
    ----------
    adata_raw : annData
        Spot x Bene raw expression matrix [S x G]

    min_perc : float
        lower-bound percentile of non-zero gexps for filtering spots

    max_perc : float
        upper-bound percentile of non-zero gexps for filtering spots

    n_top_genes: float
        number of the variable genes

    mt_thld : float
        max. percentage of mitochondrial gexps for filtering spots
        with excessive MT expressions

    multiple_data: bool
        whether the study need integrate datasets
    """
    adata = adata_raw.copy()

    if min_perc and max_perc:
        assert 0 < min_perc < max_perc < 100, \
            "Invalid thresholds for cells: {0}, {1}".format(min_perc, max_perc)
        min_counts = np.percentile(adata.obs['total_counts'], min_perc)
        sc.pp.filter_cells(adata, min_counts=min_counts)

    # Remove cells with excessive MT expressions
    # Remove MT & RB genes
    if verbose:
        LOGGER.info('Preprocessing1: delete the mt and rp')
        
    adata.var['mt'] = np.logical_or(
        adata.var_names.str.startswith('MT-'),
        adata.var_names.str.startswith('mt-')
    )
    adata.var['rb'] = adata.var_names.str.startswith(('RP', 'Rp', 'rp'))

    sc.pp.calculate_qc_metrics(adata, qc_vars=['mt'], inplace=True)
    mask_cell = adata.obs['pct_counts_mt'] < mt_thld
    mask_gene = np.logical_and(~adata.var['mt'], ~adata.var['rb'])

    adata = adata[mask_cell, mask_gene]
    sc.pp.filter_genes(adata, min_cells=1)

    # Normalize & take log-transform
    if verbose:
        LOGGER.info('Preprocessing2: Normalize')
    if multiple_data:
        sc.pp.normalize_total(adata, target_sum=1e6, inplace=True)
    else:
        sc.pp.normalize_total(adata, inplace=True)

    # Preprocessing3: Logarithm
    if verbose:
        LOGGER.info('Preprocessing3: Logarithm')
    sc.pp.log1p(adata)

    # Preprocessing4: Find the variable genes
    if verbose:
        LOGGER.info('Preprocessing4: Find the variable genes')
    sc.pp.highly_variable_genes(adata, flavor='seurat', n_top_genes=n_top_genes, inplace=True)

    # Filter corresponding `obs` & `var` in raw-count matrix
    adata_raw = adata_raw[adata.obs_names, adata.var_names]
    adata_raw.var['highly_variable'] = adata.var['highly_variable']
    adata_raw.obs = adata.obs

    return adata_raw, adata


def load_adata(data_folder, sample_id, n_genes, multiple_data=False):
    """
    load visium adata with raw counts, preprocess & extract highly variable genes

    Parameters
    ----------
        data_folder : str
            Root directory of the data

        sample_id : str
            Sample subdirectory under `data_folder`

        n_genes : int
            the number of the gene for training

        multiple_data: bool
            whether the study include multiple datasets

    Returns
    -------
        adata : sc.AnnData
            Processed ST raw counts

        adata_norm : sc.AnnData
            Processed ST normalized & log-transformed data
    """
    has_feature_h5 = os.path.isfile(
        os.path.join(data_folder, sample_id, 'filtered_feature_bc_matrix.h5')
    ) # whether dataset stored in h5 with spatial info.

    if has_feature_h5:
        adata = sc.read_visium(path=os.path.join(data_folder, sample_id), library_id=sample_id)
        adata.var_names_make_unique()
        adata.obs['sample'] = sample_id
    elif sample_id.startswith('simu'): # simulations
        adata = sc.read_csv(os.path.join(data_folder, sample_id, 'counts.st_synth.csv'))
    else:
        filenames = [
            f[:-5] for f in os.listdir(os.path.join(data_folder, sample_id))
            if f[-5:] == '.h5ad'
        ]
        assert len(filenames) == 1, \
            "None or more than `h5ad` file in the data directory," \
            "please contain only 1 target ST file in the given directory"
        adata = sc.read_h5ad(os.path.join(data_folder, sample_id, filenames[0] + '.h5ad'))
        adata.var_names_make_unique()
        adata.obs['sample'] = sample_id

    if '_index' in adata.var.columns:
        adata.var_names = adata.var['_index']
        adata.var_names.name = 'Genes'
        adata.var.drop('_index', axis=1, inplace=True)

    adata, adata_norm = preprocess(adata, n_top_genes=n_genes, multiple_data=multiple_data)
    return adata, adata_norm


def load_signatures(filename, adata):
    """
    load annotated signature gene sets

    Parameters
    ----------
    filename : str
        Signature file

    adata : sc.AnnData
        ST count matrix

    Returns
    -------
    gene_sig : pd.DataFrame
        signatures per cell type / state
    """
    assert os.path.isfile(filename), "Unable to find the signature file"
    gene_sig = pd.read_csv(filename, index_col=0)
    gene_sig = filter_gene_sig(gene_sig, adata.to_df())
    sigs = np.unique(
        gene_sig.apply(
            lambda x:
            pd.unique(x[~pd.isna(x)])
        ).values
    )

    return gene_sig, np.unique(sigs)


def preprocess_img(
    data_path,
    sample_id,
    adata_index,
    rgb_channels=True
):
    """
    Load and preprocess visium paired H&E image & spatial coords

    Parameters
    ----------
    data_path : str
        Root directory of the data

    sample_id : str
        Sample subdirectory under `data_path`

    rgb_channels : bool
        Whether to apply binary color deconvolution to extract 1D `eosin` channel
        Please refer to:
        https://digitalslidearchive.github.io/HistomicsTK/examples/color_deconvolution.html

    Returns
    -------
    img : np.ndarray
        Processed histology image

    map_info : np.ndarray
        Spatial coords of spots (dim: [S, 2])
    """
    from skimage import io
    

    filename = os.path.join(data_path, sample_id, 'spatial', 'tissue_hires_image.png')
    if os.path.isfile(filename):
        if rgb_channels:
            img = io.imread(filename)
            #img = (img-img.min())/(img.max()-img.min())
        
        else:
            try:
                import histomicstk as htk
                
            except ImportError:
                raise ImportError("Please install `histomicstk` package to process histology images\n\
                                Using `pip install histomicstk --find-links https://girder.github.io/large_image_wheels`")
            import histomicstk as htk
            img = io.imread(filename)
            if img.max() <= 1:
                img = (img * 255).astype(np.uint8)
        
            # Create stain matrix
            stains = ['hematoxylin','eosin', 'null']
            stain_cmap = htk.preprocessing.color_deconvolution.stain_color_map             
            W = np.array([stain_cmap[st] for st in stains]).T

            # Color deconvolution    
            imDeconvolved = htk.preprocessing.color_deconvolution.color_deconvolution(img, W)
            img = imDeconvolved.Stains[:,:,0]   # H-channel

            # Take inverse of H-channel (approx. cell density)      
            img = (img - img.min()) / (img.max()-img.min())       
            img = img.max() - img  
            img = (img*255).astype(np.uint8)
    else:
        img = None

    # Mapping images to location
    f = open(os.path.join(data_path, sample_id, 'spatial', 'scalefactors_json.json', ))
    json_info = json.load(f)
    f.close()

    tissue_position_list = pd.read_csv(os.path.join(data_path, sample_id, 'spatial', 'tissue_positions_list.csv'), header=None, index_col=0)
    tissue_position_list = tissue_position_list.loc[adata_index, :]
    map_info = tissue_position_list.iloc[:, -4:-2]
    map_info.columns = ['array_row', 'array_col']
    map_info.loc[:, 'imagerow'] = tissue_position_list.iloc[:, -2]
    map_info.loc[:, 'imagecol'] = tissue_position_list.iloc[:, -1]
    map_info.loc[:, 'sample'] = sample_id

    return {
        'img': img,
        'map_info': map_info,
        'scalefactor': json_info
    }


def get_adata_wsig(adata, adata_norm, gene_sig):
    """
    Select intersection of HVGs from dataset & signature annotations
    """
    # TODO: in-place operators for `adata`
    hvgs = adata.var_names[adata.var.highly_variable]
    unique_sigs = np.unique(gene_sig.values[~pd.isna(gene_sig)])
    genes_to_keep = np.union1d(
        hvgs,
        np.intersect1d(adata.var_names, unique_sigs)
    )
    return adata[:, genes_to_keep], adata_norm[:, genes_to_keep]


def filter_gene_sig(gene_sig, adata_df):
    for i in range(gene_sig.shape[0]):
        for j in range(gene_sig.shape[1]):
            gene = gene_sig.iloc[i, j]
            if gene in adata_df.columns:
                # We don't filter signature genes based on expression level (prev: threshold=20)
                if adata_df.loc[:, gene].sum() < 0:
                    gene_sig.iloc[i, j] = 'NaN'
    return gene_sig


def get_umap(adata_sample, display=False):
    sc.tl.pca(adata_sample, svd_solver='arpack')
    sc.pp.neighbors(adata_sample, n_neighbors=15, n_pcs=40)
    sc.tl.umap(adata_sample, min_dist=0.2)
    if display:
        sc.pl.umap(adata_sample)
    umap_plot = pd.DataFrame(adata_sample.obsm['X_umap'],
                             columns=['umap1', 'umap2'],
                             index=adata_sample.obs_names)
    return umap_plot


def get_simu_map_info(umap_plot):
    map_info = []
    map_info = [-umap_plot['umap2'] * 10, umap_plot['umap1'] * 10]
    map_info = pd.DataFrame(np.transpose(map_info),
                            columns=['array_row', 'array_col'],
                            index=umap_plot.index)
    return map_info


def get_windowed_library(adata_sample, map_info, library, window_size):
    library_n = []

    for i in adata_sample.obs_names:
        window_size = window_size
        dist_arr = np.sqrt(
            (map_info.loc[:, 'array_col'] - map_info.loc[i, 'array_col']) ** 2 +
            (map_info.loc[:, 'array_row'] - map_info.loc[i, 'array_row']) ** 2
        )

        library_n.append(library[dist_arr < window_size].mean())
    library_n = np.array(library_n)
    return library_n


def append_sigs(gene_sig, factor, sigs, n_genes=10):
    """
    Append list of genes to a given cell type as additional signatures or
    add novel cell type / states & their signatures
    """
    assert len(sigs) > 0, "Signature list must have positive length"
    gene_sig_new = gene_sig.copy()

    if not isinstance(sigs, list):
        sigs = sigs.to_list()
    if n_genes < len(sigs):
        sigs = sigs[:n_genes]

    markers = set([i for i in gene_sig[factor] if str(i) != 'nan'] +
                  [i for i in sigs if str(i) != 'nan'])
    nrow_diff = int(np.abs(len(markers)-gene_sig_new.shape[0]))
    if len(markers) > gene_sig_new.shape[0]:
        df_dummy = pd.DataFrame([[np.nan] * gene_sig.shape[1]] * nrow_diff, 
                                columns=gene_sig_new.columns)
        gene_sig_new = pd.concat([gene_sig_new, df_dummy], ignore_index=True)
    else:
        markers = list(markers) + [np.nan]*nrow_diff
    gene_sig_new[factor] = list(markers)

    return gene_sig_new


def refine_anchors(
    visium_args,
    aa_model,
    anchor_threshold=0.1,
    n_genes=10
):
    """
    Refine anchor spots & marker genes with archetypal analysis. We append DEGs
    computed from archetypes to their best-matched anchors followed by re-computing
    new anchor spots

    Parameters
    ----------
    visium_args : VisiumArgument
        Default parameter set for Starfysh upon dataloading

    aa_model : ArchetypalAnalysis
        Pre-computed archetype object

    anchor_threshold : float
        Top percent of anchor spots per cell-type
        for archetypal mapping

    n_genes : int
        # archetypal marker genes to append per refinement iteration

    Returns
    -------
    visimu_args : VisiumArgument
        updated parameter set for Starfysh
    """
    # TODO: integrate into `visium_args` class

    n_spots = visium_args.adata.shape[0]
    gene_sig = visium_args.gene_sig.copy()
    anchors_df = visium_args.get_anchors()
    n_top_anchors = int(anchor_threshold*n_spots)

    # Retrieve anchor-archetype mapping scores
    map_df, map_dict = aa_model.assign_archetypes(anchor_df=anchors_df[:n_top_anchors],
                                           r=n_top_anchors) 
    markers_df = aa_model.find_markers(display=False)

    # (1). Update signatures
    for cell_type, archetype in map_dict.items():
        gene_sig = append_sigs(gene_sig=gene_sig,
                               factor=cell_type,
                               sigs=markers_df[archetype],
                               n_genes=n_genes)

    # (2). Update data args.
    visium_args.gene_sig = gene_sig
    visium_args._update_anchors()
    return visium_args


# -------------------
# Post-processing
# -------------------

def extract_feature(adata, key):
    """
    Extract generative / inference output from adata.obsm
    generate dummy tmp. adata for plotting
    """
    assert key in adata.obsm.keys(), "Unfounded Starfysh generative / inference output: {}".format(key)

    if key == 'qc_m':
        cols = adata.uns['cell_types']  # cell type deconvolution
    elif key == 'qz_m':
        cols = ['z'+str(i) for i in range(adata.obsm[key].shape[1])]  # inferred qz (low-dim manifold)
    elif '_inferred_exprs' in key:
        cols = adata.var_names  # inferred cell-type specific expressions
    else:
        cols = ['density']
    adata_dummy = adata.copy()
    adata_dummy.obs = pd.DataFrame(adata.obsm[key], index=adata.obs.index, columns=cols)
    return adata_dummy


def get_reconst_img(args, img_patches):
    """
    Reconst original histology image (H x W) from the given patched image (S x P)
    """
    reconst_img = np.zeros_like(args.img, dtype=np.float64)
    r = args.params['patch_r']
    patch_size = (r * 2, r * 2, 3) if args.img.ndim == 3 else (r * 2, r * 2)
    scale_factor = args.scalefactor['tissue_hires_scalef']
    img_col = args.map_info['imagecol'] * scale_factor
    img_row = args.map_info['imagerow'] * scale_factor

    for i in range(len(img_col)):
        patch_y = slice(int(img_row[i]) - r, int(img_row[i]) + r)
        patch_x = slice(int(img_col[i]) - r, int(img_col[i]) + r)

        sy, sx = reconst_img[patch_y, patch_x].shape[:2]
        img_patch = img_patches[i].reshape(patch_size)
        reconst_img[patch_y, patch_x] = img_patch[:sy, :sx]  # edge patch cases

    return reconst_img

    



# optimization
# for parallelizing stuff
from multiprocessing import cpu_count

import numpy as np
from joblib import Parallel, delayed
from scipy.sparse import lil_matrix
from tqdm.notebook import tqdm

# get number of cores for multiprocessing
NUM_CORES = cpu_count()

##########################################################
# Helper functions for parallelizing kernel construction
##########################################################


def kth_neighbor_distance(distances, k, i):
    """Returns distance to kth nearest neighbor.

    Distances: sparse CSR matrix
    k: kth nearest neighbor
    i: index of row
    .
    """
    # convert row to 1D array
    row_as_array = distances[i, :].toarray().ravel()

    # number of nonzero elements
    num_nonzero = np.sum(row_as_array > 0)

    # argsort
    kth_neighbor_idx = np.argsort(np.argsort(-row_as_array)) == num_nonzero - k
    return np.linalg.norm(row_as_array[kth_neighbor_idx])


def rbf_for_row(G, data, median_distances, i):
    """Helper function for computing radial basis function kernel for each row of the data matrix.

    :param G: (array) KNN graph representing nearest neighbour connections between cells
    :param data: (array) data matrix between which euclidean distances are computed for RBF
    :param median_distances: (array) radius for RBF - the median distance between cell and k nearest-neighbours
    :param i: (int) data row index for which RBF is calculated
    :return: sparse matrix containing computed RBF for row
    """
    # convert row to binary numpy array
    row_as_array = G[i, :].toarray().ravel()

    # compute distances ||x - y||^2 in PC/original X space
    numerator = np.sum(np.square(data[i, :] - data), axis=1, keepdims=False)

    # compute radii - median distance is distance to kth nearest neighbor
    denominator = median_distances[i] * median_distances

    # exp
    full_row = np.exp(-numerator / denominator)

    # masked row - to contain only indices captured by G matrix
    masked_row = np.multiply(full_row, row_as_array)

    return lil_matrix(masked_row)


##########################################################
# Archetypal Analysis Metacell Graph
##########################################################


class SEACellGraph:
    """SEACell graph class."""

    def __init__(self, ad, build_on="X_pca", n_cores: int = -1, verbose: bool = False):
        """SEACell graph class.

        :param ad: (anndata.AnnData) object containing data for which metacells are computed
        :param build_on: (str) key corresponding to matrix in ad.obsm which is used to compute kernel for metacells
                        Typically 'X_pca' for scRNA or 'X_svd' for scATAC
        :param n_cores: (int) number of cores for multiprocessing. If unspecified, computed automatically as
                        number of CPU cores
        :param verbose: (bool) whether or not to suppress verbose program logging
        """
        """Initialize model parameters"""
        # data parameters
        self.n, self.d = ad.obsm[build_on].shape

        # indices of each point
        self.indices = np.array(range(self.n))

        # save data
        self.ad = ad
        self.build_on = build_on

        self.knn_graph = None
        self.sym_graph = None

        # number of cores for parallelization
        if n_cores != -1:
            self.num_cores = n_cores
        else:
            self.num_cores = NUM_CORES

        self.M = None  # similarity matrix
        self.G = None  # graph
        self.T = None  # transition matrix

        # model params
        self.verbose = verbose

    ##############################################################
    # Methods related to kernel + sim matrix construction
    ##############################################################

    def rbf(self, k: int = 15, graph_construction="union"):
        """Initialize adaptive bandwith RBF kernel (as described in C-isomap).

        :param k: (int) number of nearest neighbors for RBF kernel
        :return: (sparse matrix) constructed RBF kernel
        """
        import scanpy as sc

        if self.verbose:
            print("Computing kNN graph using scanpy NN ...")

        # compute kNN and the distance from each point to its nearest neighbors
        sc.pp.neighbors(self.ad, use_rep=self.build_on, n_neighbors=k, knn=True)
        knn_graph_distances = self.ad.obsp["distances"]

        # Binarize distances to get connectivity
        knn_graph = knn_graph_distances.copy()
        knn_graph[knn_graph != 0] = 1
        # Include self as neighbour
        knn_graph.setdiag(1)

        self.knn_graph = knn_graph
        if self.verbose:
            print("Computing radius for adaptive bandwidth kernel...")

            # compute median distance for each point amongst k-nearest neighbors
        with Parallel(n_jobs=self.num_cores, backend="threading") as parallel:
            median = k // 2
            median_distances = parallel(
                delayed(kth_neighbor_distance)(knn_graph_distances, median, i)
                for i in tqdm(range(self.n))
            )

        # convert to numpy array
        median_distances = np.array(median_distances)

        if self.verbose:
            print("Making graph symmetric...")

        print(
            f"Parameter graph_construction = {graph_construction} being used to build KNN graph..."
        )
        if graph_construction == "union":
            sym_graph = (knn_graph + knn_graph.T > 0).astype(float)
        elif graph_construction in ["intersect", "intersection"]:
            knn_graph = (knn_graph > 0).astype(float)
            sym_graph = knn_graph.multiply(knn_graph.T)
        else:
            raise ValueError(
                f"Parameter graph_construction = {graph_construction} is not valid. \
             Please select `union` or `intersection`"
            )

        self.sym_graph = sym_graph
        if self.verbose:
            print("Computing RBF kernel...")

        with Parallel(n_jobs=self.num_cores, backend="threading") as parallel:
            similarity_matrix_rows = parallel(
                delayed(rbf_for_row)(
                    sym_graph, self.ad.obsm[self.build_on], median_distances, i
                )
                for i in tqdm(range(self.n))
            )

        if self.verbose:
            print("Building similarity LIL matrix...")

        similarity_matrix = lil_matrix((self.n, self.n))
        for i in tqdm(range(self.n)):
            similarity_matrix[i] = similarity_matrix_rows[i]

        if self.verbose:
            print("Constructing CSR matrix...")

        self.M = (similarity_matrix).tocsr()
        return self.M



from . import config
from . import core
from . import presults
from . import io
from . import preprocess
from . import utils
from . import plot
from .version import __version__

__all__ = [
    "config",
    "core",
    "presults",
    "io",
    "preprocess",
    "utils",
    "plot",
    "__version__",
]



import numpy as np
import pandas as pd
import scanpy as sc
from scipy import stats
from sklearn.linear_model import Lasso, LassoCV
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from tqdm.auto import tqdm


# TF-target computation
def gene_tf_associations(
    gene_peak,
    pwm,
    min_corr=0.0,
    max_pval=0.1,
    min_peaks=0,
    ct_peaks=None,
    gene_set=None,
):
    """TODO."""
    tfs = pwm.var_names
    len(tfs)

    # initialize returns
    gene_tfs = {}
    n_gp_assoc = 0

    for gene, peak_df in tqdm(gene_peak.items(), desc="genes", total=len(gene_peak)):
        if gene_set is None:
            gene_set = gene_peak.keys()

        if gene not in gene_set or not isinstance(peak_df, pd.DataFrame):
            # skip genes with no peaks
            continue

        if ct_peaks is None:
            # metacell use case
            sig_peaks = (peak_df["pval"] < max_pval) & (peak_df["cor"] > min_corr)
            if all(~sig_peaks):
                # no significant peaks
                continue
            # grab list of sig peaks correlated with this gene
            peaks = peak_df.loc[sig_peaks, :].index
        else:
            valid_peaks = [x for x in peak_df.index if x in ct_peaks]

            if len(valid_peaks) == 0:
                continue
            peaks = peak_df.loc[valid_peaks, :].index

        if len(peaks) >= min_peaks:
            # convert pwm scores to bool, 0 = False
            tf_association = pwm[pwm.obs_names.isin(peaks), :].X
            tf_association = tf_association.toarray().astype(bool)

            # Grab TFs that are associated with a peak at least once
            tf_idxs = np.where(np.any(tf_association, axis=0))[0]

            gene_tfs[gene] = {}
            for tf in tf_idxs:
                tf_name = pwm.var_names[tf]
                tf_mask = tf_association[
                    :, tf
                ]  # subset the PWM scores to the relevant tf

                gene_tfs[gene][tf_name] = peaks[tf_mask]

                # grab the peaks for each tf
                n_gp_assoc += 1

    # log number of genes that passed thresholds
    print(
        f"{len(gene_tfs):,} genes and {n_gp_assoc:,} gene-TF combinations with at least {min_peaks} peaks."
    )

    return gene_tfs


def compute_tf_target_mat(pwm, gene_peak, gene_tfs, ct_specific=False):
    """TODO."""
    tfs = pwm.var_names

    # initiate three DFs
    tf_targ_mat = pd.DataFrame(0.0, index=gene_peak.index, columns=tfs)

    pwm_scores = pwm.to_df()

    for gene, tf_dict in tqdm(gene_tfs.items(), total=len(gene_tfs)):
        for tf, peaks in tf_dict.items():
            if ct_specific:
                score = pwm_scores.loc[peaks, tf].sum()
            else:
                # Compute weighted PWM score
                corrs = gene_peak[gene].loc[peaks, "cor"]
                score = (pwm_scores.loc[peaks, tf].values * corrs.values).sum()

            tf_targ_mat.loc[gene, tf] = score

    return tf_targ_mat


# gene selection


def get_de_genes(adata, groups, thresh, group_key, fc_min=1.5, pval_cut=1e-2):
    """TODO."""
    final_dict = {}
    print("threshold:", thresh)
    print()
    for g in tqdm(groups, total=len(groups)):
        group_dict = {}

        ref = [x for x in groups if x != g]
        _compute_de_genes(adata, g, group_dict, ref, group_key, fc_min, pval_cut)

        gene_list = []
        for gene in group_dict:
            if group_dict[gene] >= thresh:
                gene_list.append(gene)

        num_genes = len(gene_list)

        print(g)
        print("\tnum genes:", num_genes)

        gene_names = []
        for gene in gene_list:
            gene_names.append(gene)

        final_dict[g] = gene_names

    return final_dict


def _compute_de_genes(
    adata,
    g,
    group_dict,
    references,
    group_key,
    fc_min=1.5,
    pval_cut=1e-2,
):
    sc.settings.verbosity = 0
    for ref in references:
        de_data = sc.tl.rank_genes_groups(
            adata,
            groupby=group_key,
            groups=[g],
            reference=ref,
            use_raw=False,
            copy=True,
        )

        de_genes = sc.get.rank_genes_groups_df(
            de_data, g, pval_cutoff=pval_cut, log2fc_min=fc_min
        )

        valid_de_df = de_genes.sort_values(by="logfoldchanges", ascending=False)

        for gene in valid_de_df["names"]:
            if gene not in group_dict:
                group_dict[gene] = 1
            else:
                group_dict[gene] += 1
    sc.settings.verbosity = 1


def get_gene_set(
    mc_ad, peak_counts, de_genes, group_key, sub_group, expr_thresh=25, min_peaks=5
):
    """TODO."""
    mc_expr = mc_ad.to_df()
    valid_genes = peak_counts[peak_counts >= min_peaks].index

    highly_expressed = mc_expr.columns[
        (
            mc_expr.groupby(mc_ad.obs[group_key]).mean().loc[sub_group] > expr_thresh
        ).any()
    ]

    gene_set = np.intersect1d(valid_genes, de_genes)
    gene_set = np.intersect1d(gene_set, highly_expressed)

    print("Number of genes...")
    print(f"   ...highly expressed in {sub_group}: {len(highly_expressed)}")
    print(f"   ...correlated {min_peaks}+ peaks: {len(valid_genes)}")
    print(f"   ...differentially expressed: {len(de_genes)}")
    print("...taking intersection...")
    print()
    print(f"TOTAL: {len(gene_set)}")

    return gene_set


# Lasso regression


def fit_lasso_model(
    gtf, zs, tf_set, gene_set, cells, test_size=0.20, cv_fold=5, max_iter=10000
):
    """TODO."""
    X = gtf.loc[gene_set, tf_set]
    scaler = StandardScaler()
    X = pd.DataFrame(scaler.fit_transform(X), index=gene_set, columns=tf_set)

    # set up containers
    res = {}
    for key in ["coefs", "rmses", "rsqs", "sp_corr", "x_tests", "intercept", "y_stats"]:
        res[key] = {}

    y = zs.loc[gene_set, :]
    for cell in tqdm(cells, total=len(cells)):
        _train_model(cell, X, y, gene_set, res, test_size, cv_fold, max_iter)

    return res


def _train_model(cell, X, y, gene_set, res, test_size=0.20, cv_fold=5, max_iter=10000):
    y = y.loc[:, cell]
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size)

    lasso_cv_model = LassoCV(cv=cv_fold, max_iter=max_iter, n_jobs=16).fit(
        X_train, y_train
    )

    opt_alpha = lasso_cv_model.alpha_
    lasso_tuned = Lasso().set_params(alpha=opt_alpha).fit(X_train, y_train)

    y_pred = lasso_tuned.predict(X_test)

    # save stats
    rmse = np.sqrt(mean_squared_error(y_test, y_pred))
    r_sq = r2_score(y_test, y_pred)
    sp_corr = stats.spearmanr(y_test, y_pred)

    res["rmses"][cell] = rmse
    res["rsqs"][cell] = r_sq
    res["sp_corr"][cell] = sp_corr
    res["x_tests"][cell] = X_test
    res["y_stats"][cell] = y_test, y_pred
    res["intercept"][cell] = lasso_tuned.intercept_
    res["coefs"][cell] = pd.Series(lasso_tuned.coef_, index=X.columns).sort_values()


# TF-Activity functions
def non_zero_tfs(cells, ct_res, thresh=20):
    """TODO."""
    tf_count = {}

    for cell in tqdm(cells, total=len(cells)):
        coefs = ct_res["coefs"][cell]
        coefs = coefs[coefs != 0]
        for tf in coefs.index:
            if tf not in tf_count:
                tf_count[tf] = 1
            else:
                tf_count[tf] += 1
    valid_tfs = [tf for tf in tf_count.keys() if tf_count[tf] >= thresh]
    print(f"Number of Qualifying TFs: {len(valid_tfs)}")

    return valid_tfs


def compute_tf_activity(ad_sub, res, valid_tfs, cells):
    """TODO."""
    # cell_order = ad_sub.obs.loc[cells,:].sort_values('palantir_pseudotime').index
    tf_activities = pd.DataFrame(
        0.0, index=valid_tfs, columns=cells
    )  # set up container

    for cell in tqdm(cells, total=len(cells)):
        X_test = res["x_tests"][cell]
        y_test = res["y_stats"][cell][0]
        res["y_stats"][cell][1]
        coefs = res["coefs"][cell].copy()
        intercept = res["intercept"][cell]
        rmse = res["rmses"][cell]

        for tf in valid_tfs:
            if coefs.loc[tf] >= 0:
                sign = 1
            else:
                sign = -1

            new_coefs = coefs.copy()
            new_coefs.loc[tf] = 0
            new_y_pred = (X_test[new_coefs.index] * new_coefs).sum(axis=1) + intercept

            new_rmse = np.sqrt(mean_squared_error(y_test, new_y_pred))
            diff = np.abs(new_rmse - rmse)

            # fill in container
            tf_activities.loc[tf, cell] = diff * sign

    return tf_activities



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see: http://amp.pharm.mssm.edu/Enrichr/help#api for API docs

import sys, json, os, logging
from typing import Iterable, Tuple
import requests
import pandas as pd
from io import StringIO
from collections import OrderedDict
from pkg_resources import resource_filename
from time import sleep
from tempfile import TemporaryDirectory
from numpy import isscalar
from .plot import barplot
from .parser import Biomart
from .utils import *
from .stats import calc_pvalues, multiple_testing_correction
from typing import Tuple, Union, List, Dict, Iterable, Optional

class Enrichr(object):
    """Enrichr API"""
    def __init__(self, gene_list: Iterable[str], 
                 gene_sets: Union[List[str],str,Dict[str,str]], 
                 organism: str='human', 
                 descriptions: Optional[str]='',
                 outdir: Optional[str]='Enrichr', 
                 cutoff: float=0.05, 
                 background: Union[List[str],int,str] ='hsapiens_gene_ensembl',
                 format: str='pdf', 
                 figsize: Tuple[float, float] =(6.5,6), 
                 top_term: int =10, 
                 no_plot: bool=False, 
                 verbose: bool=False):

        self.gene_list = gene_list
        self.gene_sets = gene_sets
        self.descriptions = str(descriptions)
        self.outdir = outdir
        self.cutoff = cutoff
        self.format = format
        self.figsize = figsize
        self.__top_term = int(top_term)
        self.__no_plot = no_plot
        self.verbose = bool(verbose)
        self.module = "enrichr"
        self.res2d = None
        self._processes = 1
        self.background = background
        self._bg = None
        self.organism = organism
        self._organism = None
        self.ENRICHR_URL = "http://maayanlab.cloud" 
        # init logger
        logfile = self.prepare_outdir()
        self._logger = log_init(
            name=str(self.module) + str(id(self)),
            log_level=logging.INFO if self.verbose else logging.WARNING,
            filename=logfile,
        )


    def prepare_outdir(self):
        """create temp directory."""
        self._outdir = self.outdir
        if self._outdir is None:
            self._tmpdir = TemporaryDirectory()
            self.outdir = self._tmpdir.name
        elif isinstance(self.outdir, str):
            mkdirs(self.outdir)
        else:
            raise Exception("Error parsing outdir: %s"%type(self.outdir))

        # handle gene_sets
        logfile = os.path.join(self.outdir, "gseapy.%s.%s.log" % (self.module, self.descriptions))
        return logfile

    def __parse_gmt(self, g: str):
        with open(g) as genesets:
            g_dict = { line.strip().split("\t")[0]: line.strip().split("\t")[2:]
                       for line in genesets.readlines() }
        return g_dict

    def __gmt2dict(self, gene_sets: List[str]):
        """helper function, only convert gmt to dict and keep strings"""
        gss = []
        for g in gene_sets:
            # only convert gmt to dict. local mode
            if isinstance(g, str) and g.lower().endswith(".gmt"):
                if os.path.exists(g): 
                    self._logger.info("User Defined gene sets is given: %s"%g)
                    gss.append(self.__parse_gmt(g))
                else:
                    self._logger.warning("User Defined gene sets is not found: %s"%g)
            else:
                gss.append(g) 
        
        return gss 

    def parse_genesets(self):
        """parse gene_sets input file type"""

        gss = []
        if isinstance(self.gene_sets, list):
            gss = self.__gmt2dict(self.gene_sets)

        elif isinstance(self.gene_sets, str):
            gss = [ g.strip() for g in self.gene_sets.strip().split(",") ]
            gss = self.__gmt2dict(gss)


        elif isinstance(self.gene_sets, dict):
            gss = [self.gene_sets]
        else:
            raise Exception("Error parsing enrichr libraries, please provided corrected one")
        
        # now, gss[LIST] contains dict or strings.
        if len(gss) < 1:
            raise Exception("No GeneSets are valid !!! Check your gene_sets input.")
        gss_exist = [] 
        enrichr_library = []
        # if all local gmts (local mode), skip connect to enrichr server
        if not all([isinstance(g, dict) for g in gss]):
            enrichr_library = self.get_libraries()
        
        # check enrichr libraries are valid 

        for g in gss:
            if isinstance(g, dict): 
                gss_exist.append(g)
                continue
            if isinstance(g, str): 
                if g in enrichr_library: 
                    gss_exist.append(g)
                else:
                    self._logger.warning("Enrichr library not found: %s"%g)

        return gss_exist

    def parse_genelists(self):
        """parse gene list"""
        if isinstance(self.gene_list, list):
            genes = self.gene_list
        elif isinstance(self.gene_list, pd.DataFrame):
            # input type is bed file
            if self.gene_list.shape[1] >=3:
                genes= self.gene_list.iloc[:,:3].apply(lambda x: "\t".join([str(i) for i in x]), axis=1).tolist()
            # input type with weight values
            elif self.gene_list.shape[1] == 2:
               genes= self.gene_list.apply(lambda x: ",".join([str(i) for i in x]), axis=1).tolist()
            else:
               genes = self.gene_list.squeeze().tolist()
        elif isinstance(self.gene_list, pd.Series):
            genes = self.gene_list.squeeze().tolist()
        else:
            # get gene lists or bed file, or gene list with weighted values.
            genes=[]
            with open(self.gene_list) as f:
                for gene in f:
                    genes.append(gene.strip())

        self._isezid = all(map(self._is_entrez_id, genes))
        if self._isezid: 
            self._gls = set(map(int, genes))
        else:
            self._gls = genes

        return '\n'.join(genes)

    def send_genes(self, gene_list, url):
        """ send gene list to enrichr server"""
        payload = {
          'list': (None, gene_list),
          'description': (None, self.descriptions)
           }
        # response
        response = requests.post(url, files=payload, verify=True)
        if not response.ok:
            raise Exception('Error analyzing gene list')
        sleep(1)
        job_id = json.loads(response.text)

        return job_id

    def check_genes(self, gene_list, usr_list_id):
        '''
        Compare the genes sent and received to get successfully recognized genes
        '''
        response = requests.get('%s/%s/view?userListId=%s' %(self.ENRICHR_URL, self._organism, usr_list_id), verify=True)
        if not response.ok:
            raise Exception('Error getting gene list back')
        returnedL = json.loads(response.text)["genes"]
        returnedN = sum([1 for gene in gene_list if gene in returnedL])
        self._logger.info('{} genes successfully recognized by Enrichr'.format(returnedN))

    def get_results(self, gene_list):
        """Enrichr API"""
        ADDLIST_URL = '%s/%s/addList'%(self.ENRICHR_URL,  self._organism)
        job_id = self.send_genes(gene_list, ADDLIST_URL)
        user_list_id = job_id['userListId']
       
        RESULTS_URL = '%s/%s/export'%(self.ENRICHR_URL,  self._organism)
        query_string = '?userListId=%s&filename=%s&backgroundType=%s'
        # set max retries num =5
        s = retry(num=5)
        filename = "%s.%s.reports" % (self._gs, self.descriptions)
        url = RESULTS_URL + query_string % (user_list_id, filename, self._gs)
        response = s.get(url, stream=True)
        response.encoding = 'utf-8'
        if not response.ok:
            self._logger.error('Error fetching enrichment results: %s'%self._gs)

        try:
            res = pd.read_csv(StringIO(response.text), sep="\t")
        except pd.errors.ParserError as e:
            RESULTS_URL = '%s/Enrichr/enrich'%self.ENRICHR_URL
            query_string = '?userListId=%s&backgroundType=%s'
            url = RESULTS_URL + query_string % (user_list_id, self._gs)
            response = s.get(url)
            if not response.ok:
                self._logger.error('Error fetching enrichment results: %s'%self._gs)
            data = json.loads(response.text)
            colnames = ["Rank","Term", "P-value", "Z-score", "Combined Score", 
                        "Genes", "Adjusted P-value", 
                        "Old P-value", "Old adjusted P-value"]
            res = pd.DataFrame(data[self._gs], columns=colnames)
            res['Genes'] = res['Genes'].apply(";".join) 

        return [job_id['shortId'], res]

    def _is_entrez_id(self, idx):
        try:
            int(idx)
            return True
        except:
            return False
    
    def get_libraries(self):
        """return active enrichr library name. Official API """

        lib_url='%s/%s/datasetStatistics'%(self.ENRICHR_URL,  self._organism)
        response = requests.get(lib_url, verify=True)
        if not response.ok:
            raise Exception("Error getting the Enrichr libraries")
        libs_json = json.loads(response.text)
        libs = [lib['libraryName'] for lib in libs_json['statistics']]
 
        return sorted(libs)


    def get_background(self):
        """get background gene"""

        # input is a file
        if os.path.isfile(self.background):
            with open(self.background) as b:
                bg2 = b.readlines() 
            bg = [g.strip() for g in bg2]  
            return set(bg)
        
        # package included data
        DEFAULT_CACHE_PATH = os.path.join(os.path.expanduser("~"), ".cache/gseapy")
        DB_FILE = os.path.join(
            DEFAULT_CACHE_PATH, "{}.background.genes.txt".format(self.background)
        )
        if os.path.exists(DB_FILE):
            df = pd.read_csv(DB_FILE, sep="\t")
        else:
            # background is a biomart database name
            self._logger.warning(
                "Downloading %s for the first time. It might take a couple of miniutes."
                % self.background
            )
            bm = Biomart()
            df = bm.query(
                dataset=self.background,
                attributes=["ensembl_gene_id", "external_gene_name", "entrezgene_id"],
                filename=os.path.join(
                    DEFAULT_CACHE_PATH,
                    "{}.background.genes.txt".format(self.background),
                ),
            )
        self._logger.info("Using all annotated genes with GO_ID as background: %s"%self.background)
        df.dropna(subset=['entrezgene_id'], inplace=True)     
        # input id type: entrez or gene_name
        if self._isezid:
            bg = df['entrezgene_id'].astype(int)
        else:
            bg = df['external_gene_name']

        return set(bg)

    def set_organism(self):
        """Select Enrichr organism from below:
       
           Human & Mouse, H. sapiens & M. musculus
           Fly, D. melanogaster
           Yeast, S. cerevisiae
           Worm, C. elegans
           Fish, D. rerio

        """

        default = [ 'human','mouse','hs', 'mm',
                    'homo sapiens', 'mus musculus',
                    'h. sapiens', 'm. musculus']

        if self.organism.lower() in default:
            self._organism = 'Enrichr'
            return

        organism = {
                    'Fly': ['fly', 'd. melanogaster', 'drosophila melanogaster'],
                    'Yeast': ['yeast', 's. cerevisiae', 'saccharomyces cerevisiae'],
                    'Worm': ['worm', 'c. elegans', 'caenorhabditis elegans', 'nematode'],
                    'Fish': ['fish', 'd. rerio', 'danio rerio', 'zebrafish']
                  }

        for k, v in organism.items():
            if self.organism.lower() in v :
                self._organism = k+'Enrichr'
                return

        if self._organism is None:
            raise Exception("No supported organism found !!!")

        ENRICHR_SERVER = "%s/%s"%(self.ENRICHR_URL, self._organism) 
        
        if requests.get(ENRICHR_SERVER, verify=True).ok:
            return

        #self.ENRICHR_URL = 'http://amp.pharm.mssm.edu'
        ENRICHR_SERVER = "%s/%s"%(self.ENRICHR_URL, self._organism) 

        if requests.get(ENRICHR_SERVER, verify=True).ok:
            return
        else:
            raise Exception("Please check Enrichr URL is OK: %s"%self.ENRICHR_URL)

        return

    def filter_gmt(self, gmt, background):
        """the gmt values should be filtered only for genes that exist in background
           this substantially affect the significance of the test, the hypergeometric distribution.

           :param gmt: a dict of gene sets.
           :param background: list, set, or tuple. A list of custom backgound genes.  
        """
        gmt2 = {}
        for term, genes in gmt.items():
            # If value satisfies the condition, then store it in new_dict
            newgenes = [g for g in genes if g in background]
            if len(newgenes) > 0:
                gmt2[term] = newgenes 
        return gmt2

    def enrich(self, gmt):
        """use local mode
         
        p = p-value computed using the Fisher exact test (Hypergeometric test)  

        Not implemented here:

            combine score = log(p)·z

        see here: http://amp.pharm.mssm.edu/Enrichr/help#background&q=4
        
        columns contain:
            
            Term Overlap P-value Adjusted_P-value Genes

        """
        if isscalar(self.background):
            if isinstance(self.background, int) or self.background.isdigit():
                self._bg = int(self.background)
            elif isinstance(self.background, str):
                # self.background = set(reduce(lambda x,y: x+y, gmt.values(),[]))
                self._bg = self.get_background()
                self._logger.info("Background: found %s genes"%(len(self._bg)))
            else:
                raise Exception("Unsupported background data type")
        else:
            # handle array object: nd.array, list, tuple, set, Series
            try:
                it = iter(self.background)
                self._bg = set(self.background)
            except TypeError:
                self._logger.error("Unsupported background data type")
        # statistical testing
        hgtest = list(calc_pvalues(query=self._gls, gene_sets=gmt, 
                                   background=self._bg))
        if len(hgtest) > 0:
            terms, pvals, oddr, olsz, gsetsz, genes = hgtest
            fdrs, rej = multiple_testing_correction(ps = pvals, 
                                                    alpha=self.cutoff,
                                                    method='benjamini-hochberg')
            # save to a dataframe
            odict = OrderedDict()
            odict['Term'] = terms
            odict['Overlap'] = list(map(lambda h,g: "%s/%s"%(h, g), olsz, gsetsz))
            odict['P-value'] = pvals
            odict['Adjusted P-value'] = fdrs
            odict['Odds Ratio'] = oddr
            # odict['Reject (FDR< %s)'%self.cutoff ] = rej
            odict['Genes'] = [";".join(g) for g in genes]
            res = pd.DataFrame(odict)
            return res
        return 

    def run(self):
        """run enrichr for one sample gene list but multi-libraries"""

        # read input file
        genes_list = self.parse_genelists()

        self._logger.info("Connecting to Enrichr Server to get latest library names")
        gss = self.parse_genesets()
        if len(gss) < 1:
            self._logger.error("None of your input gene set matched ! %s"%self.gene_sets)
            self._logger.error("Hint: Current organism = %s, is this correct?\n"%self.organism +\
                               "Hint: use get_library_name() to view full list of supported names.")
            raise LookupError("Not validated Enrichr library ! Please provide correct organism and library name!")
        self.results = pd.DataFrame()

        for g in gss: 
            if isinstance(g, dict): 
                ## local mode
                res = self.enrich(g)
                shortID, self._gs = str(id(g)), "CUSTOM%s"%id(g)
                if res is None: 
                    self._logger.info("No hits return, for gene set: Custom%s"%shortID)
                    continue
            else:
                ## online mode
                self._gs = str(g)
                self._logger.debug("Start Enrichr using library: %s" % (self._gs))
                self._logger.info('Analysis name: %s, Enrichr Library: %s' % (self.descriptions, self._gs))
                shortID, res = self.get_results(genes_list)
                # Remember gene set library used
            res.insert(0, "Gene_set", self._gs)
            # Append to master dataframe
            #self.results = self.results.append(res, ignore_index=True)
            self.results = pd.concat([self.results, res], ignore_index=True)
            self.res2d = res
            if self._outdir is None: continue
            self._logger.info('Save file of enrichment results: Job Id:' + str(shortID))
            outfile = "%s/%s.%s.%s.reports.txt" % (self.outdir, self._gs, self.organism, self.module)
            self.res2d.to_csv(outfile, index=False, encoding='utf-8', sep="\t")
            # plotting
            if not self.__no_plot:
                msg = barplot(df=res, cutoff=self.cutoff, figsize=self.figsize,
                              top_term=self.__top_term, color='salmon',
                              title=self._gs,
                              ofname=outfile.replace("txt", self.format))
                if msg is not None : self._logger.warning(msg)
            self._logger.info('Done.\n')
        # clean up tmpdir
        if self._outdir is None: self._tmpdir.cleanup()

        return


def enrichr(gene_list, gene_sets, organism='human', description='',
            outdir='Enrichr', background='hsapiens_gene_ensembl', cutoff=0.05,
            format='pdf', figsize=(8,6), top_term=10, no_plot=False, verbose=False):
    """Enrichr API.

    :param gene_list: str, list, tuple, series, dataframe. Also support input txt file with one gene id per row. 
                      The input `identifier` should be the same type to `gene_sets`.

    :param gene_sets: str, list, tuple of Enrichr Library name(s). 
                      or custom defined gene_sets (dict, or gmt file). 
                      
                      Examples: 

                      Input Enrichr Libraries (https://maayanlab.cloud/Enrichr/#stats):
                        str: 'KEGG_2016'
                        list: ['KEGG_2016','KEGG_2013']
                        Use comma to separate each other, e.g. "KEGG_2016,huMAP,GO_Biological_Process_2018"

                      Input custom files:
                        dict: gene_sets={'A':['gene1', 'gene2',...],
                                        'B':['gene2', 'gene4',...], ...}
                        gmt: "genes.gmt"

                      see also the online docs: 
                      https://gseapy.readthedocs.io/en/latest/gseapy_example.html#2.-Enrichr-Example


    :param organism: Enrichr supported organism. Select from (human, mouse, yeast, fly, fish, worm).
                     This argument only affects the Enrichr library names you've chosen.
                     No any affects to gmt or dict input of `gene_sets`.

                     see here for more details: https://maayanlab.cloud/modEnrichr/.
                     
    :param description: optional. name of the job.
    :param outdir:   Output file directory

    :param background: int, list, str. Please ignore this argument if your input are just Enrichr library names.
                       
                       However, this argument is not straightforward when `gene_sets` is given a custom input (a gmt file or dict).
                       There are 3 ways to set this argument:
 
                       (1) (Recommended) Input a list of background genes. 
                           The background gene list is defined by your experment. e.g. the expressed genes in your RNA-seq.
                           The gene identifer in gmt/dict should be the same type to the backgound genes.  
                       
                       (2) Specify a number, e.g. the number of total expressed genes.
                           This works, but not recommend. It assumes that all your genes could be found in background.
                           If genes exist in gmt but not included in background, 
                           they will affect the significance of the statistical test.  

                       (3) (Default) Set a Biomart dataset name.
                           The background will be all annotated genes from the `BioMart datasets` you've choosen. 
                           The program will try to retrieve the background information automatically.

                           Please Use the example code below to choose the correct dataset name:
                            >>> from gseapy.parser import Biomart 
                            >>> bm = Biomart()
                            >>> datasets = bm.get_datasets(mart='ENSEMBL_MART_ENSEMBL')

    :param cutoff:   Show enriched terms which Adjusted P-value < cutoff. 
                     Only affects the output figure, not the final output file. Default: 0.05
    :param format:  Output figure format supported by matplotlib,('pdf','png','eps'...). Default: 'pdf'.
    :param figsize: Matplotlib figsize, accept a tuple or list, e.g. (width,height). Default: (6.5,6).
    :param bool no_plot: If equals to True, no figure will be drawn. Default: False.
    :param bool verbose: Increase output verbosity, print out progress of your job, Default: False.

    :return: An Enrichr object, which obj.res2d stores your last query, obj.results stores your all queries.
    
    """
    enr = Enrichr(gene_list, gene_sets, organism, description, outdir,
                  cutoff, background, format, figsize, top_term, no_plot, verbose)
    # set organism
    enr.set_organism()
    enr.run()

    return enr




from math import ceil, floor

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def plot_binarization(
    auc_mtx: pd.DataFrame, regulon_name: str, threshold: float, bins: int = 200, ax=None
) -> None:
    """
    Plot the "binarization" process for the given regulon.

    :param auc_mtx: The dataframe with the AUC values for all cells and regulons (n_cells x n_regulons).
    :param regulon_name: The name of the regulon.
    :param bins: The number of bins to use in the AUC histogram.
    :param threshold: The threshold to use for binarization.
    """
    if ax is None:
        ax = plt.gca()
    sns.distplot(auc_mtx[regulon_name], ax=ax, norm_hist=True, bins=bins)

    ylim = ax.get_ylim()
    ax.plot([threshold] * 2, ylim, "r:")
    ax.set_ylim(ylim)
    ax.set_xlabel("AUC")
    ax.set_ylabel("#")
    ax.set_title(regulon_name)


def plot_rss(rss, cell_type, top_n=5, max_n=None, ax=None):
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(4, 4))
    if max_n is None:
        max_n = rss.shape[1]
    data = rss.T[cell_type].sort_values(ascending=False)[0:max_n]
    ax.plot(np.arange(len(data)), data, ".")
    ax.set_ylim([floor(data.min() * 100.0) / 100.0, ceil(data.max() * 100.0) / 100.0])
    ax.set_ylabel("RSS")
    ax.set_xlabel("Regulon")
    ax.set_title(cell_type)
    ax.set_xticklabels([])

    font = {
        "color": "red",
        "weight": "normal",
        "size": 4,
    }

    for idx, (regulon_name, rss_val) in enumerate(
        zip(data[0:top_n].index, data[0:top_n].values)
    ):
        ax.plot([idx, idx], [rss_val, rss_val], "r.")
        ax.text(
            idx + (max_n / 25),
            rss_val,
            regulon_name,
            fontdict=font,
            horizontalalignment="left",
            verticalalignment="center",
        )



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch.autograd import Variable
from tqdm import tqdm
from torch.optim import AdamW
from torch.utils.data import Dataset, DataLoader
import pandas as pd
import numpy as np
from math import exp
import copy


class VAE(nn.Module):
    r"""
    Variational Autoencoder for bulk-to-single-cell generation.
    
    This VAE implementation is designed for learning the mapping between bulk and
    single-cell expression patterns. The model uses a beta-VAE framework with
    customizable encoder-decoder architecture for generating synthetic single cells.
    
    The architecture consists of:
    - Multi-layer encoder producing mean and variance parameters
    - Latent space sampling with reparameterization trick
    - Multi-layer decoder reconstructing single-cell expressions
    """
    
    def __init__(self, embedding_size, hidden_size_list: list, mid_hidden):
        r"""
        Initialize VAE with specified architecture.

        Arguments:
            embedding_size: Input feature dimension (number of genes)
            hidden_size_list: List of hidden layer sizes for encoder/decoder
            mid_hidden: Latent space dimensionality
            
        Returns:
            None
        """
        super(VAE, self).__init__()
        self.embedding_size = embedding_size
        self.hidden_size_list = hidden_size_list
        self.mid_hidden = mid_hidden

        self.enc_feature_size_list = [self.embedding_size] + self.hidden_size_list + [self.mid_hidden * 2]
        self.dec_feature_size_list = [self.embedding_size] + self.hidden_size_list + [self.mid_hidden]

        self.encoder = nn.ModuleList(
            [nn.Linear(self.enc_feature_size_list[i], self.enc_feature_size_list[i + 1]) for i in
             range(len(self.enc_feature_size_list) - 1)])
        self.decoder = nn.ModuleList(
            [nn.Linear(self.dec_feature_size_list[i], self.dec_feature_size_list[i - 1]) for i in
             range(len(self.dec_feature_size_list) - 1, 0, -1)])

    def encode(self, x):
        r"""
        Encode input to latent parameters.
        
        Passes input through encoder layers to produce mean and log variance
        parameters for the latent distribution.

        Arguments:
            x: Input tensor of shape (batch_size, embedding_size)
            
        Returns:
            torch.Tensor: Concatenated mean and log variance parameters
        """
        for i, layer in enumerate(self.encoder):
            x = self.encoder[i](x)
            if i != len(self.encoder) - 1:
                x = F.relu(x)
        return x

    def decode(self, x):
        r"""
        Decode latent representation to output space.
        
        Passes latent samples through decoder layers to reconstruct
        single-cell expression profiles.

        Arguments:
            x: Latent tensor of shape (batch_size, mid_hidden)
            
        Returns:
            torch.Tensor: Reconstructed expression tensor
        """
        for i, layer in enumerate(self.decoder):
            x = self.decoder[i](x)
            x = F.relu(x)
        return x

    def forward(self, x, used_device):
        r"""
        Forward pass through VAE.
        
        Performs encoding, latent sampling, decoding, and KL divergence calculation.

        Arguments:
            x: Input expression tensor
            used_device: Device for computation (CPU/GPU)
            
        Returns:
            tuple: (reconstructed_x, kl_divergence)
        """
        x = x.to(used_device)
        encoder_output = self.encode(x)
        mu, sigma = torch.chunk(encoder_output, 2, dim=1)  # mu, log_var
        hidden = torch.randn_like(sigma) + mu * torch.exp(sigma) ** 0.5  # var => std
        x_hat = self.decode(hidden)
        kl_div = 0.5 * torch.sum(torch.exp(sigma) + torch.pow(mu, 2) - 1 - sigma) / (x.shape[0] * x.shape[1])
        return x_hat, kl_div

    def get_hidden(self, x):
        r"""
        Extract latent representation from input.
        
        Encodes input and samples from the latent distribution.

        Arguments:
            x: Input expression tensor
            
        Returns:
            torch.Tensor: Sampled latent representation
        """
        encoder_output = self.encode(x)
        mu, sigma = torch.chunk(encoder_output, 2, dim=1)  # mu, log_var
        hidden = torch.randn_like(sigma) * torch.exp(sigma) ** 0.5 + mu  # var => std
        return hidden


# bulk deconvolution
class BulkDataset(Dataset):
    r"""
    PyTorch Dataset for single-cell data with labels.
    
    Wraps single-cell expression data and corresponding labels for use
    with PyTorch DataLoader in VAE training.
    """
    
    def __init__(self, single_cell, label):
        r"""
        Initialize dataset with expression data and labels.

        Arguments:
            single_cell: Single-cell expression matrix
            label: Cell type labels
            
        Returns:
            None
        """
        self.sc = single_cell
        self.label = label

    def __getitem__(self, idx):
        r"""
        Get single item from dataset.

        Arguments:
            idx: Index of item to retrieve
            
        Returns:
            tuple: (expression_data, label)
        """
        tmp_x = self.sc[idx]
        tmp_y_tag = self.label[idx]

        return (tmp_x, tmp_y_tag) 

    def __len__(self):
        r"""
        Get dataset size.
        
        Returns:
            int: Number of samples in dataset
        """
        return self.label.shape[0]


def train_vae(single_cell, label, used_device, batch_size, feature_size, epoch_num, learning_rate, hidden_size, patience=10):
    r"""
    Train VAE model for bulk-to-single-cell generation.
    
    Trains a beta-VAE model using single-cell reference data with early stopping
    based on reconstruction loss. The model learns to generate realistic single-cell
    expression profiles.

    Arguments:
        single_cell: Single-cell expression matrix (cells x genes)
        label: Cell type labels for single cells
        used_device: Computation device (CPU/GPU)
        batch_size: Training batch size
        feature_size: Number of input features (genes)
        epoch_num: Maximum training epochs
        learning_rate: Optimizer learning rate
        hidden_size: Latent space dimensionality
        patience: Early stopping patience (10)

    Returns:
        tuple: (best_model, training_history)
    """
    batch_size = batch_size
    feature_size = feature_size
    epoch_num = epoch_num
    lr = learning_rate
    hidden_list = [2048, 1024, 512]

    mid_hidden_size = hidden_size
    weight_decay = 5e-4
    dataloader = DataLoader(BulkDataset(single_cell=single_cell, label=label), batch_size=batch_size, shuffle=True,
                            pin_memory=True)
    criterion = nn.MSELoss()
    beta = 4
    beta1 = 0.9
    beta2 = 0.999
    vae = VAE(feature_size, hidden_list, mid_hidden_size).to(used_device)
    optimizer = AdamW(vae.parameters(), lr=lr, weight_decay=weight_decay, betas=(beta1, beta2))

    pbar = tqdm(range(epoch_num))
    min_loss = 1000000000000000
    vae.train()
    early_stop = 0
    history = []

    for epoch in pbar:
        train_loss = 0

        for batch_idx, data in enumerate(dataloader):
            cell_feature, label = data
            cell_feature = torch.tensor(cell_feature, dtype=torch.float32).to(used_device)
            x_recon, total_kld = vae(cell_feature, used_device)

            recon_loss = criterion(x_recon, cell_feature)
            beta_vae_loss = recon_loss + beta * total_kld
            loss = beta_vae_loss

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            train_loss += loss.item()

        if train_loss < min_loss:
            min_loss = train_loss
            best_vae = copy.deepcopy(vae)
            early_stop = 0
            epoch_final = epoch  
        else:
            early_stop += 1

        history.append(train_loss)
        if early_stop == patience:
            print('Early stopping at epoch {}...'.format(epoch+1))
            break
        pbar.set_description('Train Epoch: {}'.format(epoch))
        pbar.set_postfix(loss=f"{train_loss:.4f}", min_loss=f"{min_loss:.4f}")
    print(f"min loss = {min_loss}")
    
    return best_vae,history


def load_vae(feature_size, hidden_size, path, used_device):
    r"""
    Load pre-trained VAE model from file.
    
    Loads a previously trained VAE model with specified architecture.

    Arguments:
        feature_size: Number of input features (genes)
        hidden_size: Latent space dimensionality
        path: Path to saved model state dict
        used_device: Computation device (CPU/GPU)

    Returns:
        VAE: Loaded VAE model
    """
    hidden_list = [2048, 1024, 512]
    vae = VAE(feature_size, hidden_list, hidden_size).to(used_device)
    vae.load_state_dict(torch.load(path, map_location=used_device))
    
    return vae


def generate_vae(net, ratio, single_cell, label, breed_2_list, index_2_gene, cell_number_target_num=None,
                 used_device=None):
    r"""
    Generate synthetic single-cell data using trained VAE.
    
    Uses a trained VAE model to generate synthetic single-cell expression profiles
    matching specified cell-type target numbers.

    Arguments:
        net: Trained VAE model
        ratio: Generation ratio (usually 1)
        single_cell: Reference single-cell expression matrix
        label: Cell type labels for reference cells
        breed_2_list: List mapping label indices to cell type names
        index_2_gene: List of gene names
        cell_number_target_num: Target number of cells per type (None)
        used_device: Computation device (CPU/GPU) (None)

    Returns:
        tuple: (generated_metadata, generated_expression_data)
    """
    # net in cuda now
    for p in net.parameters():  # reset requires_grad
        p.requires_grad = False  # avoid computation

    net.eval()
    net.to(used_device)
    cell_all_generate = []
    label_all_generate = []

    all_to_generate = 0
    for x in cell_number_target_num.values():
        all_to_generate += x

    if cell_number_target_num != None:
        epochs = 10000 
        ratio = 1
    else:
        epochs = 1

    cell_feature = torch.from_numpy(single_cell).float()
    label = torch.from_numpy(label)

    with torch.no_grad():
        with tqdm(total=all_to_generate, desc='generating') as pbar:
            for epoch in range(epochs):
                key_list = []  # list
                generate_num = 0

                label_list = label.tolist()
                for i in range(len(label_list)):
                    if cell_number_target_num[label_list[i]] <= 0:
                        continue
                    else:
                        cell_number_target_num[label_list[i]] -= 1
                        generate_num += 1
                        key_list.append(i)

                if cell_number_target_num == None or all_to_generate == 0 or len(key_list) == 0:
                    assert all_to_generate == 0 and len(key_list) == 0
                    break

                import random
                random.shuffle(key_list)

                label = label.index_select(0, torch.tensor(key_list))
                cell_feature = cell_feature.index_select(0, torch.tensor(key_list))

                dataloader = DataLoader(BulkDataset(single_cell=cell_feature, label=label), batch_size=300,
                                        shuffle=False,
                                        pin_memory=True, num_workers=0)
                for batch_idx, data in enumerate(dataloader):  
                    cell_feature_batch, label_batch = data
                    cell_feature_batch = cell_feature_batch.to(used_device)

                    label_batch = label_batch.cpu().numpy()

                    for j in range(ratio): 
                        ans_l, _ = net(cell_feature_batch, used_device)
                        ans_l = ans_l.cpu().data.numpy()
                        cell_all_generate.extend(ans_l)
                        label_all_generate.extend(label_batch)

                all_to_generate -= generate_num
                pbar.update(generate_num)

    print("generated done!")
    generate_sc_meta, generate_sc_data = prepare_data(cell_all_generate, label_all_generate, breed_2_list,
                                                      index_2_gene)
    return generate_sc_meta, generate_sc_data


def prepare_data(cell_all_generate, label_all_generate, breed_2_list, index_2_gene):
    r"""
    Format generated data into structured DataFrames.
    
    Converts generated expression arrays and labels into properly formatted
    DataFrames with cell and gene annotations.

    Arguments:
        cell_all_generate: Generated expression matrix
        label_all_generate: Generated cell type labels
        breed_2_list: List mapping indices to cell type names
        index_2_gene: List of gene names

    Returns:
        tuple: (metadata_dataframe, expression_dataframe)
    """
    cell_all_generate = np.array(cell_all_generate)
    label_all_generate = np.array(label_all_generate)

    cell_all_generate_csv = pd.DataFrame(cell_all_generate)
    label_all_generate_csv = pd.DataFrame(label_all_generate)

    ids = label_all_generate_csv[0].tolist()
    breeds = []
    for id in ids:
        breeds.append(breed_2_list[id])
    name = ["C_" + str(i + 1) for i in range(label_all_generate.shape[0])]

    label_all_generate_csv.insert(1, "Cell_type", np.array(breeds))
    label_all_generate_csv.insert(1, "Cell", np.array(name))
    label_all_generate_csv = label_all_generate_csv.drop([0], axis=1)

    cell_all_generate_csv = cell_all_generate_csv.T
    cell_all_generate_csv.columns = name
    cell_all_generate_csv.index = index_2_gene

    return label_all_generate_csv, cell_all_generate_csv



import networkx as nx
import numpy as np
import pandas as pd
import scipy, random
import scanpy as sc
from tqdm import tqdm
import matplotlib.pyplot as plt
from sklearn.metrics import pairwise_distances_chunked, pairwise_distances
from .visualize import link_plot

def coords2adjacentmat(coords,output_mode = 'adjacent',strategy_t = 'convex'):
    if strategy_t == 'convex': ### slow but may generate more reasonable delaunay graph
        from libpysal.cg import voronoi_frames
        from libpysal import weights
        cells, _ = voronoi_frames(coords, clip="convex hull")
        delaunay_graph = weights.Rook.from_dataframe(cells).to_networkx()
    elif strategy_t == 'delaunay': ### fast but may generate long distance edges
        from scipy.spatial import Delaunay
        from collections import defaultdict
        tri = Delaunay(coords)
        delaunay_graph = nx.Graph()
        coords_dict = defaultdict(list)
        for i, coord in enumerate(coords):
            coords_dict[tuple(coord)].append(i)
        for simplex in tri.simplices:
            for i in range(3):
                for node1 in coords_dict[tuple(coords[simplex[i]])]:
                    for node2 in coords_dict[tuple(coords[simplex[(i+1)%3]])]:
                        if not delaunay_graph.has_edge(node1, node2):
                            delaunay_graph.add_edge(node1, node2)
    if output_mode == 'adjacent':
        return nx.to_scipy_sparse_array(delaunay_graph).todense()
    elif output_mode == 'raw':
        return delaunay_graph
    elif output_mode == 'adjacent_sparse':
        return nx.to_scipy_sparse_array(delaunay_graph)

def hv_cutoff(max_col,threshold=2000):
    for thres_t in range(0,int(np.max(max_col))):
        if np.sum(max_col > thres_t) < threshold:
            return thres_t -1

def detect_highly_variable_genes(sdata,batch_key = 'batch',n_top_genes = 4000,count_layer = 'count'):
    samples = np.unique(sdata.obs[batch_key])
    thres_list = []
    max_count_list = []
    bool_list = []
    for list_t, sample_t in enumerate(samples):
        idx_t = sdata.obs[batch_key] == sample_t
        if count_layer == '.X':
            max_t = sdata[idx_t,:].X.max(0).toarray() if scipy.sparse.issparse(sdata.X) else sdata[idx_t,:].X.max(0)
        else:
            max_t = sdata[idx_t,:].layers[count_layer].max(0).toarray() if scipy.sparse.issparse(sdata.layers[count_layer]) else sdata[idx_t,:].layers[count_layer].max(0)
        max_count_list.append(max_t)
        thres_list.append(hv_cutoff(max_count_list[list_t],threshold=n_top_genes))
        bool_list.append(max_count_list[list_t] > thres_list[list_t])
    stack = np.stack(bool_list)
    return np.all(stack, axis=0)[0]

def extract_coords_exp(sdata, batch_key = 'batch', cols = 'spatial', count_layer = 'count', data_format = 'norm1e4',ifcombat = False, if_inte = False):
    coords_raw = {}
    exps = {}
    samples = np.unique(sdata.obs[batch_key])
    if count_layer == '.X':
        sdata.layers['raw'] = sdata.X.copy()
    sdata = preprocess_fast(sdata, mode = 'customized')
    if if_inte:
        scaled_layer = 'log2_norm1e4_scaled'
        pc_feature = 'X_pca_harmony'
        sdata = Harmony_integration(sdata,
                                    scaled_layer = scaled_layer,
                                    use_highly_variable_t = True,
                                    batch_key = batch_key,
                                    umap_n_neighbors = 50,
                                    umap_n_pcs = 30,
                                    min_dist = 0.01,
                                    spread_t = 5,
                                    source_sample_ctype_col = None,
                                    output_path = None,
                                    n_components = 50,
                                    ifplot = False,
                                    ifcombat = ifcombat)
        for sample_t in samples:
            idx_t = sdata.obs[batch_key] == sample_t
            coords_raw[sample_t] = sdata.obsm['spatial'][idx_t] if type(cols) is not list else np.array(sdata.obs[cols][idx_t])
            exps[sample_t] = sdata[idx_t].obsm[pc_feature].copy()
    else:
        sdata.X = sdata.layers[data_format].copy()
        if ifcombat == True:
            sc.pp.combat(sdata, key=batch_key)
        for sample_t in samples:
            idx_t = sdata.obs[batch_key] == sample_t
            coords_raw[sample_t] = sdata.obsm['spatial'][idx_t] if type(cols) is not list else np.array(sdata.obs[cols][idx_t])
            exps[sample_t] = sdata[idx_t].X.copy()
            if scipy.sparse.issparse(exps[sample_t]):
                exps[sample_t] = exps[sample_t].toarray()
    return coords_raw,exps

def Harmony_integration(
    sdata_inte,
    scaled_layer,
    use_highly_variable_t,
    batch_key,
    umap_n_neighbors,
    umap_n_pcs,
    min_dist,
    spread_t,
    source_sample_ctype_col,
    output_path,
    n_components = 50,
    ifplot = True,
    ifcombat = False):
    #### integration based on the Harmony
    sdata_inte.X = sdata_inte.layers[scaled_layer].copy()
    if ifcombat == True:
        sc.pp.combat(sdata_inte, key=batch_key)
    print(f'Running PCA based on the layer {scaled_layer}:')
    sc.tl.pca(sdata_inte, use_highly_variable=use_highly_variable_t, svd_solver = 'full', n_comps= n_components)
    print(f'Running Harmony integration:')
    sc.external.pp.harmony_integrate(sdata_inte, batch_key)
    print(f'Compute a neighborhood graph based on the {umap_n_neighbors} `n_neighbors`, {umap_n_pcs} `n_pcs`:')
    sc.pp.neighbors(sdata_inte, n_neighbors=umap_n_neighbors, n_pcs=umap_n_pcs, use_rep='X_pca_harmony')
    print(f'Generate the UMAP based on the {min_dist} `min_dist`, {spread_t} `spread`:')
    sc.tl.umap(sdata_inte,min_dist=min_dist, spread = spread_t)
    sdata_inte.obsm['har_X_umap'] = sdata_inte.obsm['X_umap'].copy()
    if ifplot == True:
        plt.rcParams.update({'pdf.fonttype':42})
        sc.settings.figdir = output_path
        sc.set_figure_params(figsize=(10, 10),facecolor='white',vector_friendly=True, dpi_save=300,fontsize = 25)
        sc.pl.umap(sdata_inte,color=[batch_key],size=10,save=f'_har_{umap_n_pcs}pcs_batch.pdf')
        sc.pl.umap(sdata_inte,color=[source_sample_ctype_col],size=10,save=f'_har_{umap_n_pcs}pcs_ctype.pdf') if source_sample_ctype_col is not None else None
    return sdata_inte

def random_sample(coords_t, nodenum, seed_t = 2):
    random.seed(seed_t)
    sub_node_idx = np.sort(random.sample(range(coords_t.shape[0]),nodenum))
    return sub_node_idx

def sub_node_sum(coords_t,exp_t,nodenum=1000,vis = True,seed_t = 2):
    from scipy.sparse import csr_matrix as csr
    # random.seed(seed_t)
    if nodenum > coords_t.shape[0]:
        print('The number of nodes is larger than the total number of nodes. Return the original data.')
        sub_node_idx = np.arange(coords_t.shape[0])
        if scipy.sparse.issparse(exp_t):
            return exp_t,sub_node_idx
        else:
            return csr(exp_t),sub_node_idx
    # sub_node_idx = np.sort(random.sample(range(coords_t.shape[0]),nodenum))
    sub_node_idx = random_sample(coords_t, nodenum, seed_t = seed_t)
    coords_t_sub = coords_t[sub_node_idx,:].copy()
    close_idx = nearest_neighbors_idx(coords_t_sub,coords_t)
    A = np.zeros([coords_t_sub.shape[0],coords_t.shape[0]])
    for ind,i in enumerate(close_idx.tolist()):
        A[i,ind] = 1
    csr_A = csr(A)
    if scipy.sparse.issparse(exp_t):
        exp_t_sub = csr_A.dot(exp_t)
    else:
        exp_t_sub = csr_A.dot(csr(exp_t))
    if(vis == True):
        link_plot(close_idx,coords_t,coords_t_sub,k = 1)
    return exp_t_sub,sub_node_idx

def nearest_neighbors_idx(coord1,coord2,mode_t = 'knn'): ### coord1 is the reference, coord2 is the target
    if mode_t == 'knn':
        from sklearn.neighbors import KNeighborsClassifier
        knn_classifier = KNeighborsClassifier(n_neighbors=1,metric='euclidean')
        knn_classifier.fit(coord1, np.zeros(coord1.shape[0]))  # Use dummy labels, since we only care about distances
        # Find nearest neighbors
        _, close_idx = knn_classifier.kneighbors(coord2)
        return close_idx
    else:
        result = []
        dists = pairwise_distances_chunked(coord2,coord1,working_memory = 100, metric='euclidean', n_jobs=-1)
        for chunk in tqdm(dists): # for each chunk (minibatch)
            knn_ind = np.argpartition(chunk, 0, axis=-1)[:, 0] # introsort to get indices of top k neighbors according to the distance matrix [n_query, k]
            result.append(knn_ind)
        close_idx = np.concatenate(result)
        return np.expand_dims(close_idx, axis=1)
    
def non_zero_center_scale(sdata_t_X):
    std_nocenter = np.sqrt(np.square(sdata_t_X).sum(0)/(sdata_t_X.shape[0]-1))
    return(sdata_t_X/std_nocenter)

def sub_data_extract(sample_list,coords_raw, exps, nodenum_t = 20000, if_non_zero_center_scale = True):
    coords_sub = dict()
    exp_sub = dict()
    sub_node_idxs = dict()
    for sample_t in sample_list:
        exp_t,sub_node_idxs[sample_t] = sub_node_sum(coords_raw[sample_t],exps[sample_t],nodenum=nodenum_t,vis = False)
        exp_sub[sample_t] = non_zero_center_scale(exp_t.toarray()) if if_non_zero_center_scale else exp_t.toarray()
        coords_sub[sample_t] = coords_raw[sample_t][sub_node_idxs[sample_t],:]
    return coords_sub,exp_sub,sub_node_idxs

def preprocess_fast(sdata1, mode = 'customized',target_sum=1e4,base = 2,zero_center = True,regressout = False):
    print('Preprocessing...')
    from scipy.sparse import csr_matrix as csr
    if 'raw' in sdata1.layers:
        if type(sdata1.layers['raw']) != scipy.sparse._csr.csr_matrix:
            sdata1.layers['raw'] = csr(sdata1.layers['raw'].copy())
        sdata1.X = sdata1.layers['raw'].copy()
    else:
        if type(sdata1.X) != scipy.sparse._csr.csr_matrix:
            sdata1.X = csr(sdata1.X.copy())
        sdata1.layers['raw'] = sdata1.X.copy()
    if mode == 'default':
        sc.pp.normalize_total(sdata1)
        sdata1.layers['norm'] = csr(sdata1.X.copy())
        sc.pp.log1p(sdata1)
        sdata1.layers['log1p_norm'] = csr(sdata1.X.copy())
        sc.pp.scale(sdata1,zero_center = zero_center)
        if scipy.sparse.issparse(sdata1.X): #### automatically change to non csr matrix (zero_center == True, the .X would be sparce)
            sdata1.X = sdata1.X.toarray().copy()
        sdata1.layers['log1p_norm_scaled'] = sdata1.X.copy()
        if regressout:
            sdata1.obs['total_counts'] = sdata1.layers['raw'].toarray().sum(axis=1)
            sc.pp.regress_out(sdata1, ['total_counts'])
            sdata1.layers['log1p_norm_scaled'] = sdata1.X.copy()
        return sdata1 #### sdata1.X is sdata1.layers['log1p_norm_scaled']
    elif mode == 'customized':
        if target_sum == 1e4:
            target_sum_str = '1e4'
        else:
            target_sum_str = str(target_sum)
        sc.pp.normalize_total(sdata1,target_sum=target_sum)
        sdata1.layers[f'norm{target_sum_str}'] = csr(sdata1.X.copy())
        sc.pp.log1p(sdata1,base = base)
        sdata1.layers[f'log{str(base)}_norm{target_sum_str}'] = csr(sdata1.X.copy())
        sc.pp.scale(sdata1,zero_center = zero_center)
        if scipy.sparse.issparse(sdata1.X): #### automatically change to non csr matrix (zero_center == True, the .X would be sparce)
            sdata1.X = sdata1.X.toarray().copy()
        sdata1.layers[f'log{str(base)}_norm{target_sum_str}_scaled'] = sdata1.X.copy()
        if regressout:
            sdata1.obs['total_counts'] = sdata1.layers['raw'].toarray().sum(axis=1)
            sc.pp.regress_out(sdata1, ['total_counts'])
            sdata1.layers[f'log{str(base)}_norm{target_sum_str}_scaled'] = sdata1.X.copy()
        return sdata1 #### sdata1.X is sdata1.layers[f'log{str(base)}_norm{target_sum_str}_scaled']
    else:
        print('Please set the `mode` as one of the {"default", "customized"}.')

def cell_select(coords_t, s=0.5, c=None, output_path_t=None):
    '''
    Select cells by drawing a polygon on the plot.
    Click the "Finish Polygon" button to finish drawing the polygon.
    Click the "Clear Polygon" button to clear the polygon.
    '''
    import matplotlib.pyplot as plt
    from matplotlib.patches import Polygon
    from shapely.geometry import Point, Polygon as ShapelyPolygon
    import ipywidgets as widgets
    import numpy as np

    indices = np.arange(coords_t.shape[0]).reshape(-1, 1)
    coords = np.hstack((coords_t, indices))
    
    global poly_coords, polygon_patch, selected_cell_ids
    poly_coords = []
    polygon_patch = None
    selected_cell_ids = []

    def process_selected_ids(selected_ids):
        # Inner function to process the selected cell IDs
        print("Selected Cell IDs:", selected_ids)
        # Additional processing can be done here

    def on_click(event):
        global poly_coords, polygon_patch
        if event.inaxes is not ax or event.button != 1:
            return
        poly_coords.append((event.xdata, event.ydata))
        if polygon_patch:
            polygon_patch.remove()
        polygon_patch = Polygon(poly_coords, closed=False, color='blue', alpha=0.3)
        ax.add_patch(polygon_patch)
        fig.canvas.draw()

    def finish_polygon(b):
        global poly_coords, polygon_patch, selected_cell_ids
        if polygon_patch:
            polygon_patch.set_closed(True)
            fig.canvas.draw()
            shapely_poly = ShapelyPolygon(poly_coords)
            selected_cell_ids = [int(id) for x, y, id in coords if shapely_poly.contains(Point(x, y))]
            process_selected_ids(selected_cell_ids)
            if output_path_t is not None:
                fig.savefig(output_path_t)
            poly_coords.clear()

    def clear_polygon(b):
        global poly_coords, polygon_patch
        if polygon_patch:
            polygon_patch.remove()
            polygon_patch = None
            fig.canvas.draw()
            poly_coords.clear()

    fig, ax = plt.subplots(figsize=[10, 10])
    x, y, ids = zip(*coords)
    ax.scatter(x, y, s=s, c=c)
    ax.set_aspect('equal', adjustable='box')

    finish_button = widgets.Button(description="Finish Polygon")
    finish_button.on_click(finish_polygon)

    clear_button = widgets.Button(description="Clear Polygon")
    clear_button.on_click(clear_polygon)

    display(widgets.HBox([finish_button, clear_button]))

    fig.canvas.mpl_connect('button_press_event', on_click)

#### Delta analysis


def get_neighborhood_rad(coords_centroids, coords_candidate, radius_px, dist=None):
    if dist is None:
        dist =  pairwise_distances(coords_centroids, coords_candidate, metric='euclidean', n_jobs=-1)
    rad_mask = dist < radius_px
    return rad_mask

def delta_cell_cal(coords_tgt,coords_ref,ctype_tgt,ctype_ref,radius_px):
    '''
    coords_tgt: coordinates of niche centroids (target cells).
    coords_ref: coordinates of reference cells.
    ctype_tgt: cell type of niche centroids.
    ctype_ref: cell type of reference cells.
    radius_px: radius of neighborhood.
    
    Output:
    return: delta_cell_tgt, delta_cell_ref, delta_cell.

    e.g.
    coords_tgt = coords_final['injured']
    coords_ref = coords_final['normal']
    ctype_tgt = sdata.obs['Annotation'][right_idx]
    ctype_ref = sdata.obs['Annotation'][left_idx]
    radius_px = 1000
    df_delta_cell_tgt,df_delta_cell_ref,df_delta_cell = delta_cell(coords_tgt,coords_ref,ctype_tgt,ctype_ref,radius_px)
    '''
    ##### 1. generate nbhd_mask_tgt and nbhd_mask_ref.
    # nbhd_mask_tgt: coords_tgt vs coords_tgt itself.
    # nbhd_mask_ref: coords_tgt vs coords_ref.
    nbhd_mask_tgt = get_neighborhood_rad(coords_tgt, coords_tgt, radius_px)
    nbhd_mask_ref = get_neighborhood_rad(coords_tgt, coords_ref, radius_px)
    
    ##### 2. generate ctype_one_hot_array
    # ctype_one_hot_array: one-hot encoding of cell types.
    # To make the order of columns consistent, we stack the two labels together.
    ctype_all = np.hstack([ctype_tgt,ctype_ref])
    idx_ctype_tgt = np.arange(len(ctype_tgt))
    idx_ctype_ref = np.arange(len(ctype_tgt),len(ctype_tgt)+len(ctype_ref))
    ctype_one_hot = pd.get_dummies(ctype_all)
    ctype_one_hot_cols = ctype_one_hot.columns
    ctype_one_hot_tgt = ctype_one_hot.values[idx_ctype_tgt]
    ctype_one_hot_ref = ctype_one_hot.values[idx_ctype_ref]

    ##### 3. generate delta_cell_tgt, delta_cell_ref and delta_cell
    # delta_cell_tgt: raw cell type counts of target cells given niche centroids.
    # delta_cell_ref: raw cell type counts of reference cells given niche centroids.
    # delta_cell: delta_cell_tgt - delta_cell_ref.
    d_cell_tgt = nbhd_mask_tgt.astype(int).dot(ctype_one_hot_tgt.astype(int))
    d_cell_ref = nbhd_mask_ref.astype(int).dot(ctype_one_hot_ref.astype(int))
    d_cell = d_cell_tgt - d_cell_ref

    return pd.DataFrame(d_cell_tgt,columns = ctype_one_hot_cols), pd.DataFrame(d_cell_ref,columns = ctype_one_hot_cols), pd.DataFrame(d_cell,columns = ctype_one_hot_cols)

def delta_exp_cal(coords_tgt,coords_ref,exp_tgt,exp_ref,radius_px,valid_tgt_idx=None,valid_ref_idx=None):
    '''
    coords_tgt: coordinates of niche centroids (target cells).
    coords_ref: coordinates of reference cells.
    exp_tgt: gene expression of target cells.
    exp_ref: gene expression of reference cells.
    radius_px: radius of neighborhood.
    
    Output:
    return: delta_exp_tgt, delta_exp_ref, delta_exp.

    e.g.
    '''
    ##### 0. generate valid_tgt_idx and valid_ref_idx. For the cell type specific analysis, we only consider the cells in the given cell type.
    valid_tgt_idx = np.arange(len(coords_tgt)) if valid_tgt_idx is None else valid_tgt_idx
    valid_ref_idx = np.arange(len(coords_ref)) if valid_ref_idx is None else valid_ref_idx

    ##### 1. generate nbhd_mask_tgt and nbhd_mask_ref.
    # nbhd_mask_tgt: coords_tgt vs coords_tgt itself.
    # nbhd_mask_ref: coords_tgt vs coords_ref.
    nbhd_mask_tgt = get_neighborhood_rad(coords_tgt, coords_tgt[valid_tgt_idx], radius_px)
    nbhd_mask_ref = get_neighborhood_rad(coords_tgt, coords_ref[valid_ref_idx], radius_px)

    ##### 2. generate delta_cell_tgt, delta_cell_ref and delta_cell
    # delta_exp_tgt: Average gene expression of target cells given niche centroids.
    # delta_exp_ref: Average gene expression of reference cells given niche centroids.
    # delta_exp: delta_exp_tgt - delta_exp_ref.
    d_exp_tgt = nbhd_mask_tgt.dot(exp_tgt[valid_tgt_idx]).astype(float) / nbhd_mask_tgt.sum(axis=1)[:,None]
    d_exp_ref = nbhd_mask_ref.dot(exp_ref[valid_ref_idx]).astype(float) / nbhd_mask_ref.sum(axis=1)[:,None]

    ### if the nbhd_mask_tgt.sum(axis=1)[:,None] is 0, then the d_exp_tgt, d_exp_ref will be nan. We set it to 0.
    d_exp_tgt[np.isnan(d_exp_tgt)] = 0 
    d_exp_ref[np.isnan(d_exp_ref)] = 0
    d_exp = d_exp_tgt - d_exp_ref

    return d_exp_tgt, d_exp_ref, d_exp

def delta_exp_sigplot(p_values,avg_differences,abs_10logp_cutoff = None, abs_avg_diff_cutoff = None, sig = True):
    y_t = np.array(-np.log10(p_values))
    x_t = np.array(avg_differences)
    abs_10logp_cutoff = np.quantile(np.abs(y_t),0.95) if abs_10logp_cutoff is None else abs_10logp_cutoff
    abs_avg_diff_cutoff = np.quantile(np.abs(x_t),0.95) if abs_avg_diff_cutoff is None else abs_avg_diff_cutoff
    idx_sig = (np.abs(y_t) > abs_10logp_cutoff) & (np.abs(x_t) > abs_avg_diff_cutoff) if sig else np.zeros(len(y_t),dtype = bool)
    idx_sig_up = (y_t > abs_10logp_cutoff) & (x_t > abs_avg_diff_cutoff) if sig else np.zeros(len(y_t),dtype = bool)
    idx_sig_down = (y_t > abs_10logp_cutoff) & (x_t < -abs_avg_diff_cutoff) if sig else np.zeros(len(y_t),dtype = bool)
    plt.figure(figsize = (10,10))
    plt.scatter(x_t,y_t, s = 2, c = 'black', rasterized = True)
    plt.scatter(x_t[idx_sig],y_t[idx_sig], s = 5, c = 'red', rasterized = True)
    plt.xlabel('Average difference')
    plt.ylabel('-log10(p)')
    return idx_sig, idx_sig_up, idx_sig_down

def delta_exp_statistics(delta_exp_tgt, delta_exp_ref):
    from scipy.stats import ranksums
    from tqdm import tqdm
    p_values = []
    avg_differences = []
    for i in tqdm(range(delta_exp_tgt.shape[1])):
        # Calculate the rank-sum p-value
        p_value = ranksums(delta_exp_tgt[:, i], delta_exp_ref[:, i]).pvalue
        p_values.append(p_value)
        # Calculate the average of the differences
        avg_difference = np.mean(delta_exp_tgt[:, i] - delta_exp_ref[:, i])
        avg_differences.append(avg_difference)
    return p_values, avg_differences


"""
Functions for preprocessing of single cell RNA-seq counts
"""

import numpy as np
import anndata
from scipy.sparse import issparse


def filter_counts_data(data, cell_min_molecules=1000, genes_min_cells=10):
    """Remove low molecule count cells and low detection genes

    :param data: Counts matrix: Cells x Genes
    :param cell_min_molecules: Minimum number of molecules per cell
    :param genes_min_cells: Minimum number of cells in which a gene is detected
    :return: Filtered counts matrix
    """

    # Molecule and cell counts
    ms = data.sum(axis=1)
    cs = data.sum()

    # Filter
    return data.loc[ms.index[ms > cell_min_molecules], cs.index[cs > genes_min_cells]]


def normalize_counts(data):
    """Correct the counts for molecule count variability

    :param data: Counts matrix: Cells x Genes
    :return: Normalized matrix
    """
    ms = data.sum(axis=1)
    norm_df = data.div(ms, axis=0).mul(np.median(ms), axis=0)
    return norm_df


def log_transform(data, pseudo_count=0.1):
    """Log transform the matrix

    :param data: Counts matrix: Cells x Genes or Anndata object
    :return: Log transformed matrix
    """
    if isinstance(data, anndata.AnnData):
        if issparse(data.X):
            data.X.data = np.log2(data.X.data + pseudo_count) - np.log2(pseudo_count)
        else:
            data.X = np.log2(data.X + pseudo_count) - np.log2(pseudo_count)
    else:
        return np.log2(data + pseudo_count)


import numpy as np
import scipy as s
from .basic_distributions import Distribution
from .bernoulli import Bernoulli
from .univariate_gaussian import UnivariateGaussian

from ..utils import *

class BernoulliGaussian(Distribution):
    """
    Class to define a Bernoulli-Gaussian distributions (for more information see Titsias and Gredilla, 2014)

    The best way to characterise a joint Bernoulli-Gaussian distribution P(N,B) is by considering
    its factorisation p(N|B)p(B) where B is a bernoulli distribution and N|B=0 and N|B=1 are normal distributions

    Equations:
    p(N,B) = Normal(N|mean,var) * Bernoulli(B|theta)
    FINISH EQUATIONS

    ROOM FOR IMPROVEMENT: i think the current code is inefficient because you have to keep track of the params
    and expectations in both the factorised distributions and the joint one. I think
    perhaps is better not to define new Bernoulli and UnivariateGaussian but work directly with the joint model
    """
    def __init__(self, dim, mean_B0, mean_B1, var_B0, var_B1, theta, EN_B0=None, EN_B1=None, EB=None):
        Distribution.__init__(self,dim)
        self.B = Bernoulli(dim=dim, theta=theta, E=EB)
        self.N_B0 = UnivariateGaussian(dim=dim, mean=mean_B0, var=var_B0, E=EN_B0)
        self.N_B1 = UnivariateGaussian(dim=dim, mean=mean_B1, var=var_B1, E=EN_B1)

        # Collect parameters
        self.params = { 'mean_B0':self.N_B0.params['mean'],
                        'mean_B1':self.N_B1.params['mean'],
                        'var_B0':self.N_B0.params['var'],
                        'var_B1':self.N_B1.params['var'],
                        'theta':self.B.params['theta'] }

        # Collect expectations
        self.updateExpectations()

    def getParameters(self):
        # Get function for parameters
        return self.params

    def setParameters(self,**params):
        # Setter function for parameters
        self.B.setParameters(theta=params['theta'])
        self.N_B0.setParameters(mean=params['mean_B0'], var=params['var_B0'])
        self.N_B1.setParameters(mean=params['mean_B1'], var=params['var_B1'])
        self.params = params

    def updateParameters(self):
        # Method to update the parameters of the joint distribution based on its constituent distributions
        self.params = { 'theta':self.B.params["theta"],
                        'mean_B0':self.N_B0.params["mean"], 'var_B0':self.N_B0.params["var"],
                        'mean_B1':self.N_B1.params["mean"], 'var_B1':self.N_B1.params["var"] }

    def updateExpectations(self):
        # Method to calculate the expectations based on the current estimates for the parameters

        # Update expectations of the constituent distributions
        self.B.updateExpectations()
        self.N_B0.updateExpectations()
        self.N_B1.updateExpectations()

        # Calculate expectations of the joint distribution
        EB = self.B.getExpectation()
        EN = self.N_B1.getExpectation()
        E = EB * EN
        # TODO double check the order here 
        E2 = EB * (np.square(EN) + self.params["var_B1"])
        ENN = EB*(np.square(EN)+self.params["var_B1"]) + (1-EB)*self.params["var_B0"]

        # Compute the expectation of X*X.T (where X=BN)
        # Work but not useful now !
        # TODO : remove the loop below
        #EXXT = np.zeros((self.dim[0], self.dim[1], self.dim[1]))
        #for n in range(self.dim[0]):
        #    EXXT[n, :, :] = np.dot(E[n, :].T, E[n, :])
        #    var_n = E2[n,:] - np.square(E[n,:])
        #    EXXT[n, :, :] += np.diag(var_n)
        #

        # Collect expectations
        self.expectations = {'E': E, 'EB': EB, 'EN': EN, 'E2': E2, 'ENN': ENN}
        #self.expectations = {'E':E, 'EB':EB, 'EN':EN, 'E2':E2, 'ENN':ENN, 'EXXT':EXXT }

    def removeDimensions(self, axis, idx):

        # Method to remove undesired dimensions
        # - axis (int): axis from where to remove the elements
        # - idx (numpy array): indices of the elements to remove
        assert axis <= len(self.dim)
        assert np.all(idx < self.dim[axis])
        self.B.removeDimensions(axis,idx)
        self.N_B0.removeDimensions(axis,idx)
        self.N_B1.removeDimensions(axis,idx)

        # TODO : check this add
        dim = list(self.dim)
        dim[1] -= len(idx)
        self.dim = tuple(dim)

        self.updateParameters()
        self.updateExpectations()

    def updateDim(self, axis, new_dim):
        # Function to update the dimensionality of a particular axis
        self.B.updateDim(axis,new_dim)
        self.N_B0.updateDim(axis,new_dim)
        self.N_B1.updateDim(axis,new_dim)
        dim = list(self.dim)
        dim[axis] = new_dim
        self.dim = tuple(dim)



r"""
PopV: Population-level Variation analysis for single-cell data.

PopV is a framework for analyzing population-level variation in single-cell data,
focusing on automated cell type annotation, batch integration, and algorithm 
benchmarking. It provides a unified interface for comparing and evaluating
different single-cell analysis methods.

Key features:
- Automated cell type annotation using multiple algorithms
- Comprehensive batch integration methods
- Algorithm benchmarking and evaluation metrics
- Hub for pre-trained models and reference datasets
- Population-level variation analysis
- Integration with major single-cell frameworks

Main modules:
    algorithms: Collection of annotation and integration algorithms
    annotation: Cell type annotation utilities and workflows
    hub: Model hub for pre-trained models and reference data
    preprocessing: Data preprocessing and quality control
    visualization: Plotting and visualization functions

Supported algorithms:
- CellTypist: Automated cell type annotation
- SCANVI: Semi-supervised annotation with variational inference
- Harmony: Fast integration of single-cell data
- SCVI: Single-cell variational inference
- OnClass: Ontology-guided cell type annotation
- And many more through the algorithms module

Examples:
    >>> import omicverse as ov
    >>> # Automated annotation
    >>> ov.popv.annotation.annotate_cell_types(
    ...     adata, 
    ...     model='best_human_model'
    ... )
    >>> 
    >>> # Batch integration
    >>> ov.popv.algorithms.harmony_integrate(
    ...     adata,
    ...     batch_key='batch'
    ... )
    >>> 
    >>> # Model evaluation
    >>> metrics = ov.popv.hub.evaluate_model(
    ...     adata,
    ...     model_name='my_model'
    ... )

Notes:
    - Requires Python 3.10+ for pre-trained model compatibility
    - Some models are hosted on HuggingFace and require internet access
    - Integrates seamlessly with scanpy and other single-cell tools
"""

import sys
import warnings

if sys.version_info[:2] != (3, 10):
    warnings.warn(
        "Pretrained models on huggingface are trained with Python 3.11. "
        f"Detected Python {sys.version.split()[0]} will not load these models.",
        UserWarning,
        stacklevel=3,
    )

# Set default logging handler to avoid logging with logging.lastResort logger.
import logging

import scanpy as sc

from ._settings import settings  # isort: skip

# Import order to avoid circular imports
from . import algorithms, annotation, hub, preprocessing, visualization
from ._utils import create_ontology_resources

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata
package_name = "popv"
__version__ = importlib_metadata.version(package_name)

settings.verbosity = logging.INFO

# Jax sets the root logger, this prevents double output.
popv_logger = logging.getLogger("popv")
popv_logger.propagate = False


__all__ = [
    "algorithms",
    "annotation",
    "create_ontology_resources",
    "hub",
    "preprocessing",
    "settings",
    "visualization",
]



import numpy as np
import pandas as pd
from sklearn.linear_model import Lasso
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

import scipy.stats as stat
import math
import os
import scipy
import time
import scipy.sparse as sp
from scipy.stats import pearsonr




def readexp(expression_txt):
    f = open(expression_txt)
    gene = []
    gene_exp = {}
    cellname = []
    AA = []
    flag = 0
    for p in f:
        t = p.split()
        flag += 1
        if flag == 1:
            cellname = t
            # cellnum = len(t)
            continue
        gene.append(t[0])
        gene_exp[t[0]] = [float(t[i]) for i in range(1, len(t))]
        tt = list(map(float, t[1:]))
        AA.append(tt)
    f.close()
    
    return AA,gene_exp,cellname


def readlabel(labelname):

    type_cell = {}
    f = open(labelname)
    for p in f:
        t = p.split('\t')
        t[1] = t[1].split('\n')[0]
        if t[1] not in type_cell.keys():
            type_cell[t[1]] = []
        type_cell[t[1]].append(t[0])
    f.close()
    
    f = open(labelname)
    label = []
    labels = set()
    for p in f:
        t = p.split()
        label.append(t[1])
        labels.add(str(t[1]))
    labelsl = list(labels)
    label_index = {}
    for i in labelsl:
        label_index[i] = []
    flag = 0
    for i in label:
        flag += 1
        label_index[i].append(flag - 1)
    f.close()
    
    return labelsl,label_index,type_cell



def LassoRegression(degree, alpha):
    return Pipeline([
        ("poly", PolynomialFeatures(degree=degree)),
        ("std_scaler", StandardScaler()),
        ("lasso_reg", Lasso(alpha=alpha))])


def lasso_regress(gene_1, gene_2, degree, alpha):

    le_g1 = len(gene_1)
    t = np.linspace(min(gene_1), max(gene_1), num=le_g1)
    a1, b1 = (np.array(x).reshape(-1, 1) for x in zip(*sorted(zip(gene_1, gene_2), key=lambda x: x[0])))
    lasso1_reg = LassoRegression(degree, alpha)
    lasso1_reg.fit(a1, t)
    gene1 = lasso1_reg.predict(a1)
    gene2 = lasso1_reg.predict(b1)

    return gene1, gene2


def rss_calculate(model, right_v, left1_v, left2_v):

    model.fit(left1_v, right_v)
    rssu_1 = np.sum((model.predict(left1_v) - right_v) ** 2)
    model.fit(left2_v, right_v)
    rssr_1 = np.sum((model.predict(left2_v) - right_v) ** 2)

    return rssu_1, rssr_1


def granger(gene1, gene2, cellnum):
    import statsmodels.api as sm
    sort_1 = list(sorted(zip(gene1, gene2), key=lambda x: x[0]))

    a1, b1 = (list(x) for x in zip(*sort_1))
    a1_cha = ((300 * pd.Series(a1)) / 90001).tolist()
    b1_cha = ((300 * pd.Series(b1)) / 90001).tolist()

    model = LinearRegression()
    b1_t_1 = np.array(b1_cha[0:cellnum - 1])
    a1_t1_t2_b1_1 = sm.add_constant(np.array(pd.concat([pd.Series(a1_cha),
                                                        pd.Series(b1_cha)], axis=1)[1:]))
    b1_t1_t2_1 = sm.add_constant(np.array(b1_cha[1:cellnum]))
    b1_rssu_1, b1_rssr_1 = rss_calculate(model, b1_t_1, a1_t1_t2_b1_1, b1_t1_t2_1)
    f1_1 = ((b1_rssr_1 - b1_rssu_1) / 1) / (b1_rssu_1 / (cellnum - 3))
    p1_1 = scipy.stats.f.sf(f1_1, 1, cellnum - 3)

    b1_t = np.array(b1_cha[0:cellnum - 2])
    a1_t1_t2_b1 = sm.add_constant(np.array(pd.concat([pd.Series(a1_cha)[1:cellnum - 1].reset_index(drop=True),
                                                      pd.Series(a1_cha)[2:cellnum].reset_index(drop=True),
                                                      pd.Series(b1_cha)[1:cellnum - 1].reset_index(drop=True),
                                                      pd.Series(b1_cha)[2:cellnum].reset_index(drop=True)], axis=1)))
    b1_t1_t2 = sm.add_constant(np.array(pd.concat([pd.Series(b1_cha)[1:cellnum - 1].reset_index(drop=True),
                                                   pd.Series(b1_cha)[2:cellnum].reset_index(drop=True)], axis=1)))
    b1_rssu_2, b1_rssr_2 = rss_calculate(model, b1_t, a1_t1_t2_b1, b1_t1_t2)
    f1_2 = ((b1_rssr_2 - b1_rssu_2) / 2) / (b1_rssu_2 / (cellnum - 2 - 2 - 1))
    p1_2 = scipy.stats.f.sf(f1_2, 2, cellnum - 2 - 2 - 1)

    b1_t_3 = np.array(b1_cha[0:cellnum - 3])
    a1_t1_t2_b1_3 = sm.add_constant(np.array(pd.concat([pd.Series(a1_cha)[1:cellnum - 2].reset_index(drop=True),
                                                        pd.Series(a1_cha)[3:cellnum].reset_index(drop=True),
                                                        pd.Series(a1_cha)[2:cellnum - 1].reset_index(drop=True),
                                                        pd.Series(b1_cha)[1:cellnum - 2].reset_index(drop=True),
                                                        pd.Series(b1_cha)[3:cellnum].reset_index(drop=True),
                                                        pd.Series(b1_cha)[2:cellnum - 1].reset_index(drop=True)],
                                                       axis=1)))
    b1_t1_t2_3 = sm.add_constant(np.array(pd.concat([pd.Series(b1_cha)[1:cellnum - 2].reset_index(drop=True),
                                                     pd.Series(b1_cha)[3:cellnum].reset_index(drop=True),
                                                     pd.Series(b1_cha)[2:cellnum - 1].reset_index(drop=True)], axis=1)))
    b1_rssu_3, b1_rssr_3 = rss_calculate(model, b1_t_3, a1_t1_t2_b1_3, b1_t1_t2_3)
    f1_3 = ((b1_rssr_3 - b1_rssu_3) / 3) / (b1_rssu_3 / (cellnum - 7))
    p1_3 = scipy.stats.f.sf(f1_3, 3, cellnum - 7)

    p_1 = min([p1_1, p1_2, p1_3])

    return p_1




def granger_(gene1, gene2, cellnum):
    import statsmodels.api as sm
    from sklearn.linear_model import LinearRegression
    from statsmodels.tsa.stattools import grangercausalitytests
    # Convert gene lists to NumPy array and sort by gene1
    genes = np.array(sorted(zip(gene1, gene2), key=lambda x: x[0]))
    a1_cha, b1_cha = (300 * genes / 90001).T  # Vectorized scaling

    # Initialize the LinearRegression model
    model = LinearRegression()
    
    # Prepare the variables and pre-compute as much as possible
    p_values = []
    for lag in range(1, 4):
        lag_slice = slice(lag, cellnum)
        reset_index = lambda series: series.reset_index(drop=True)
        
        # Previous values (up to lag) for regression calculation
        b1_t = b1_cha[:-lag]
        
        # Lagged values of a1 and b1
        lagged_a1 = [reset_index(pd.Series(a1_cha)[i:cellnum - lag]) for i in range(lag)]
        lagged_b1 = [reset_index(pd.Series(b1_cha)[i:cellnum - lag]) for i in range(lag)]

        # Combine arrays for the regression
        a1_t1_t2_b1 = sm.add_constant(np.column_stack(lagged_a1 + lagged_b1))
        b1_t1_t2 = sm.add_constant(np.column_stack(lagged_b1))
        
        # Calculate RSS for unrestricted and restricted models
        rssu, rssr = rss_calculate(model, b1_t, a1_t1_t2_b1, b1_t1_t2)

        # Compute the F-statistic
        f_stat = ((rssr - rssu) / lag) / (rssu / (cellnum - lag * 2 - 1))

        # Compute the p-value for the F-statistic
        p_value = scipy.stats.f.sf(f_stat, lag, cellnum - lag * 2 - 1)
        p_values.append(p_value)

    # Return the minimum p-value from the tests for each lag
    return min(p_values)




def calculate_block(spaceblocki, gene, AA, gene_exp, ax_nl, cellnum, gene_num,pear):
    i1, i2, j1, j2 = [int(spaceblocki[i]) for i in range(len(spaceblocki))]
   # print(i1,i2,j1,j2)
    #PP = sp.dok_matrix((gene_num,gene_num))
    # Convert to more efficient sparse matrix format for arithmetics
    PP = sp.lil_matrix((gene_num, gene_num))

    # Precompute the maximum non-one value in pear multiplied by 0.3
    pear_max_multiplied = np.max(np.abs(pear - np.eye(*pear.shape)), axis=1) * 0.3
    
    for i in range(i1, i2):
        for j in range(j1, j2):
            if i == j:
                continue
            

            """
            entry_11 = gene[i]
            entry_21 = gene[j]
            mp = list(pear[i])
            if 1 in pear[i]:
                mp.remove(1)
            mp = abs(np.array(mp))
            mpmax = max(mp)
            if abs(stat.pearsonr(AA[i],AA[j])[0]) < mpmax * 0.3:
                continue
            if abs(pear[i,j]) < mpmax * 0.3:
                continue
            """
            # Skip calculation when condition is not met
            if abs(pear[i, j]) < pear_max_multiplied[i]:
                continue

            Gene1, Gene2 = lasso_regress(gene_exp[gene[i]], gene_exp[gene[j]], 30, ax_nl)

            if (Gene1 == Gene2).all() == True:
                continue

            p = granger(Gene1, Gene2, cellnum)
           # print(i,j,p)

            Pi = gene.index(gene[i])
            Pj = gene.index(gene[j])
            PP[Pi,Pj] += p
            # Update the PP matrix
            #PP[i, j] += p

    return PP.todok()


def celltype_score(labelsl, label_index, M):

    llen = len(labelsl) 
    N = np.zeros((llen, llen))
    for l1 in labelsl:
        I = label_index[l1]
        for l2 in labelsl:
            J = label_index[l2]
            if l1 == l2:
                for i in I:
                    for j in J:
                        N[labelsl.index(l1)][labelsl.index(l2)] += abs(M[i][j])
            if l1 != l2:
                for i in I:
                    for j in J:
                        if M[i][j] >= 0:
                            N[labelsl.index(l1)][labelsl.index(l2)] += M[i][j]
                        if M[i][j] < 0:
                            N[labelsl.index(l2)][labelsl.index(l1)] += abs(M[i][j])
    Nn = np.zeros((llen, llen))
    for i in range(len(N)):
        for j in range(len(N)):
            if N[i][j] == 0:
                continue
            Nn[i][j] += math.log(N[i][j], 10)

    return Nn



def GRN_GNIPLR(AA,gene_exp,step,process):
    
        print(time.asctime()+ ': Start calculating the GRN adjacency matrix.')
    
        
        gene_num = len(gene_exp)
        gene = list(gene_exp.keys())
        cellnum = len(AA[0])
    
        '''
        Pear = np.zeros((len(AA),len(AA)))
        for i in range(len(AA)):
            for j in range(i,len(AA)):
                Pear[i][j] += stat.pearsonr(AA[i], AA[j])[0]
        Pear = np.where(Pear,Pear,Pear.T)
        '''
        
        Pear = np.corrcoef(AA)
        print(time.asctime()+ ': Pearson correlation coefficient calculation completed.')
    
        space = list(np.linspace(start=0, stop=step * int(gene_num / step), num=int(gene_num / step) + 1))
        if (gene_num / step == int(gene_num / step)) == False:
            space.append(gene_num)
        space1 = []
        for i in range(len(space) - 1):
            space1.append((space[i], space[i + 1]))
        space_block = []
        for i in space1:
            for j in space1:
                space_block.append(i + j)
        print(time.asctime()+ ': GRN subprocesses start: ',len(space_block))     
        import multiprocessing
        pool = multiprocessing.Pool(process)
    
        GRN = sp.dok_matrix((gene_num, gene_num))
    
        rest = []
        from tqdm import tqdm
        for si in tqdm(space_block):
            rest.append(pool.apply_async(func=calculate_block, args=(si, gene, AA, gene_exp, 0.1, cellnum, gene_num,Pear,)))

        print(time.asctime()+ ': Waiting for all GRN subprocesses done...')
        for r in tqdm(rest):
            GRN+=r.get()
    
        print(time.asctime()+ ': Waiting for all subprocesses done...')
        pool.close()
        pool.join()
    
        print(time.asctime()+ ": Complete the GRN calculation.")

        GRN = GRN.toarray()
        
        return GRN



def MDIC3_score(AA, GRN,labelsl, label_index):
    
    begin=time.asctime()
    
    print(begin+ ': Start calculating cell-cell communication.')

    U1, S1, V1 = np.linalg.svd(AA)
    shapeA = np.shape(AA)
    d = min(shapeA) - len(S1)
    dd = []
    for i in range(d):
        dd.append(0)
    S11 = np.append(S1, dd)

    S11 = np.diag(S11)
    if shapeA[0] < shapeA[1]:
        b = np.zeros((shapeA[0], shapeA[1] - shapeA[0]))
        S11 = np.column_stack((S11, b))
    if shapeA[0] > shapeA[1]:
        b = np.zeros((shapeA[0] - shapeA[1], shapeA[1]))
        S11 = np.row_stack((S11, b))

    T = np.dot(GRN, S11)
    Tp = np.linalg.pinv(T)
    ccc_adjacency = np.dot(Tp, AA)
    type_adjacency = celltype_score(labelsl, label_index, ccc_adjacency)
    print(time.asctime()+ ": Complete the cell-cell communication calculation.")

    return ccc_adjacency,type_adjacency
    
    
    
def MDIC3_scoresave(ccc_adjacency,type_adjacency,labelsl):

    print('Saving cell-cell communication results...')
    
    fw = open('celltype_communication.txt', 'w')
    for tt in range(len(labelsl)):
        if tt < len(labelsl) - 1:
            fw.write(labelsl[tt] + '\t')
        if tt == len(labelsl) - 1:
            fw.write(labelsl[tt] + '\n')
    for i in range(len(type_adjacency)):
        Ni = list(type_adjacency[i])
        fw.write(str(labelsl[i]) + '\t')
        for ni in range(len(Ni)):
            if ni < len(Ni) - 1:
                fw.write(str(Ni[ni]) + '\t')
            if ni == len(Ni) - 1:
                fw.write(str(Ni[ni]) + '\n')
    fw.close()
    
    np.savetxt('cellular_communication.txt', ccc_adjacency)
    
    
    





def readLRDB(LRDB_txt):
    L = []
    R = []
    LR = []
    f = open(LRDB_txt)
    flag = 0
    for p in f:
        flag += 1
        if flag == 1:
            continue
        t1 = p.split('\n')[0]
        LR.append(t1)
        tlr = t1.split(' - ')
        tlr[0] = tlr[0].split(' ')[0]
        l = []
        r = []
        l.append(tlr[0])
        L.append(l)
        if tlr[1][0] == '(':
            tr = tlr[1].split('+')
            r.append(tr[0][1:])
            r.append(tr[1][:-1])
            R.append(r)
        else:
            r.append(tlr[1])
            R.append(r)

    f.close()
    
    return L,R,LR



def check_zero_majority(nums):
    count_zeros = sum(1 for num in nums if num == 0)

    return count_zeros > len(nums) * 0.7


def pearson(x, y):
    if np.std(x) == 0 or np.std(y) == 0:
        return 0
    else:
        corr, p = pearsonr(x, y)
        return corr, p


def FDRP(lrcorr):
    
    from statsmodels.stats.multitest import multipletests

    p_values = [v[1] for v in lrcorr.values()]
    corrected_p_values = multipletests(p_values, method='bonferroni')[1]
    p_LR_corr = {}
    index = 0
    for key, value in lrcorr.items():
        p_LR_corr[key] = [value[0], corrected_p_values[index]]
        index += 1
    p2_LR_corr = {key: value for key, value in p_LR_corr.items() if value[1] <= 0.01}

    return p2_LR_corr



def LR_exp(LR,L,R,geneexp,type_cell,cellname):
    
    celltype = list(type_cell.keys())
    typeindex = {}
    for i in celltype:
        cell0 = type_cell[i]
        cell0index = []
        for j in cell0:
            cell0index.append(cellname.index(j))
        typeindex[i] = cell0index
    
    gene2exp = {}
    for i in range(len(LR)):
    
        if len(L[i]) == 1 and L[i][0] not in geneexp.keys():
            continue
    
        if len(R[i]) == 1 and R[i][0] not in geneexp.keys():
            continue
    
        if len(R[i]) > 1:
            for r in R[i]:
                r0 = R[i][0]
                r1 = R[i][1]
            if r0 not in geneexp.keys():
                continue
            if r1 not in geneexp.keys():
                continue
    
        if len(L[i]) == 1:
            lexp = geneexp[L[i][0]]
            gene2exp[L[i][0]] = geneexp[L[i][0]]
    
        if len(R[i]) == 1:
            rexp = geneexp[R[i][0]]
            gene2exp[R[i][0]] = geneexp[R[i][0]]
    
        if len(R[i]) > 1:
            for r in R[i]:
                r0 = R[i][0]
                r1 = R[i][1]
            gene2exp[r0] = geneexp[r0]
            gene2exp[r1] = geneexp[r1]
    
    geneexp2 = {}
    for k in gene2exp.keys():
        n = 0
        exp = gene2exp[k]
        kexp = []
        for tt in celltype:
            for t in typeindex[tt]:
                kexp.append(exp[t])
            count = sum(1 for num in kexp if num > 0)
            if count > len(kexp) * 1 / 2:
                n += 1
        if n >= len(type_cell) / 2:
            continue
        geneexp2[k] = gene2exp[k]
    
    LRexp = {}
    for i in range(len(LR)):
        if len(L[i]) == 1 and L[i][0] not in geneexp2.keys():
            continue
        if len(R[i]) == 1 and R[i][0] not in geneexp2.keys():
            continue
        if len(R[i]) > 1:
            for r in R[i]:
                r0 = R[i][0]
                r1 = R[i][1]
            if r0 not in geneexp2.keys():
                continue
            if r1 not in geneexp2.keys():
                continue
        LRexp[LR[i]] = []
        if len(L[i]) == 1:
            lexp = geneexp2[L[i][0]]
        if len(R[i]) == 1:
            rexp = geneexp2[R[i][0]]
        if len(R[i]) > 1:
            for r in R[i]:
                r0 = R[i][0]
                r1 = R[i][1]
            r01 = geneexp2[r0] + geneexp2[r1]
            rexp = [x / 2 for x in r01]
        LRexp[LR[i]].append(lexp)
        LRexp[LR[i]].append(rexp)
    
    return LRexp
    
   
def MDIC3_SortLR(L,R,LR,target,type_cell,cellname,geneexp):
    
    
    LRexp = LR_exp(LR,L,R,geneexp,type_cell,cellname)
    cell0 = type_cell[target[0]]
    cell0index = []
    for i in cell0:
        cell0index.append(cellname.index(i))
    cell1 = type_cell[target[1]]
    cell1index = []
    for i in cell1:
        cell1index.append(cellname.index(i))

    targetexp = {}
    for key, value in LRexp.items():
        new_lists = []

        list0 = [value[0][i] for i in cell0index]
        list1 = [value[1][i] for i in cell1index]

        list0.sort(reverse=True)
        list1.sort(reverse=True)

        result0 = check_zero_majority(list0)
        result1 = check_zero_majority(list1)

        if result0 == True:
            continue
        if result1 == True:
            continue

        if len(list0) < len(list1):
            list1 = list1[:len(list0)]
        else:
            list0 = list0[:len(list1)]

        new_lists.append(list0)
        new_lists.append(list1)

        targetexp[key] = new_lists

    LRcorr = {}
    LRPcorr = {}
    for key, value in targetexp.items():
        x = value[0]
        y = value[1]
        if np.std(x) == 0 or np.std(y) == 0:
            continue
        correlation, p = pearsonr(x, y)
        LRPcorr[key] = [correlation, p]
        if p > 0.05:
            continue
        LRcorr[key] = [correlation, p]

    p2_LR_corr = FDRP(LRcorr)
    from operator import itemgetter
    sorted_LRcorr = dict(sorted(p2_LR_corr.items(), key=itemgetter(1), reverse=True))
    

    return sorted_LRcorr



def MDIC3_LRsave(sorted_LRcorr):

    print('Saving LR identification results...')

    fw = open('target_LR.txt', 'w')
    fw.write('L_R_pair' + '\t' + 'pearson_corr' + '\t' + 'P_value' + '\n')
    for k in sorted_LRcorr.keys():
        fw.write(str(k) + '\t')
        kvalue = sorted_LRcorr[k]
        fw.write(str(kvalue[0]) + '\t' + str(kvalue[1]) + '\n')
    fw.close()



class pyMDIC3(object):
    
    def __init__(self,adata,clusters):
        AA1=adata.X.T
        from scipy.sparse import issparse, csr_matrix
        if issparse(AA1):
            AA1=AA1.toarray()
        gene_exp1={}
        for gene in adata.var_names:
            gene_exp1[gene]=adata[:,gene].to_df().values.reshape(-1).tolist()
        cellname1=adata.obs_names.tolist()
        
        labels1=list(set(adata.obs[clusters]))
        label_index1={}
        for idx,cellobs in enumerate(adata.obs[clusters].tolist()):
            if cellobs not in label_index1.keys():
                label_index1[cellobs]=[idx]
            else:
                label_index1[cellobs].append(idx)
        label_cell1={}
        for cellobs in labels1:
            label_cell1[cellobs]=adata.obs.loc[adata.obs[clusters]==cellobs].index.tolist()
            
        self.adata=adata
        self.AA=AA1
        self.gene_exp=gene_exp1
        self.cellname=cellname1
        self.labels=labels1
        self.label_index=label_index1
        self.label_cell=label_cell1
            
    def GRN_train(self,step=15,process=4):
        GRN = GRN_GNIPLR(self.AA, self.gene_exp, step, process)
        self.GRN=GRN
        return GRN
    
    def CCC_cal(self,):
        # Infer the cell-cell communication
        ccc_adjacency, type_adjacency = MDIC3_score(self.AA, self.GRN, self.labels, self.label_index)
        return ccc_adjacency, type_adjacency


'''
Copy from cellrank

'''
from typing import Any, Dict, Type, Tuple, Literal, Callable

from abc import ABC, ABCMeta
from enum import Enum, EnumMeta
from functools import wraps

_DEFAULT_BACKEND = "loky"
Backend_t = Literal["loky", "multiprocessing", "threading"]


class PrettyEnum(Enum):
    r"""Enum with a pretty __str__ and __repr__."""

    @property
    def v(self) -> Any:
        r"""Alias for value attribute."""
        return self.value

    def __repr__(self) -> str:
        return f"{self.value!r}"

    def __str__(self) -> str:
        return f"{self.value!s}"


def _pretty_raise_enum(cls: Type["ErrorFormatterABC"], func: Callable) -> Callable:
    r"""Decorator to provide pretty error messages for enum classes.
    
    Arguments:
        cls: Enum class type
        func: Function to wrap
        
    Returns:
        Wrapped function with improved error formatting
    """
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> "ErrorFormatterABC":
        try:
            return func(*args, **kwargs)
        except ValueError as e:
            _cls, value, *_ = args
            e.args = (cls._format(value),)
            raise e

    if not issubclass(cls, ErrorFormatterABC):
        raise TypeError(f"Class `{cls}` must be subtype of `ErrorFormatterABC`.")
    elif not len(cls.__members__):
        # empty enum, for class hierarchy
        return func

    return wrapper


class ABCEnumMeta(EnumMeta, ABCMeta):
    r"""Metaclass for abstract enum classes."""
    def __call__(cls, *args, **kwargs):  # noqa
        if getattr(cls, "__error_format__", None) is None:
            raise TypeError(
                f"Can't instantiate class `{cls.__name__}` "
                f"without `__error_format__` class attribute."
            )
        return super().__call__(*args, **kwargs)

    def __new__(  # noqa: D102
        cls, clsname: str, superclasses: Tuple[type], attributedict: Dict[str, Any]
    ):
        res = super().__new__(cls, clsname, superclasses, attributedict)
        res.__new__ = _pretty_raise_enum(res, res.__new__)
        return res


class ErrorFormatterABC(ABC):
    r"""Abstract base class for error formatting in enums."""
    __error_format__ = "Invalid option `{!r}` for `{}`. Valid options are: `{}`."

    @classmethod
    def _format(cls, value) -> str:
        r"""Format error message for invalid enum value.
        
        Arguments:
            value: Invalid value that was provided
            
        Returns:
            Formatted error message string
        """
        return cls.__error_format__.format(
            value, cls.__name__, [m.value for m in cls.__members__.values()]
        )


class ModeEnum(str, ErrorFormatterABC, PrettyEnum, metaclass=ABCEnumMeta):
    r"""String-based enum with error formatting capabilities."""
    def _generate_next_value_(self, start, count, last_values):
        r"""Generate next value for auto() enum members.
        
        Arguments:
            start: Starting value
            count: Current count
            last_values: Previously generated values
            
        Returns:
            Lowercase string representation
        """
        return str(self).lower()


import numpy as np
import pandas as pd
import anndata as ad
import scanpy as sc
import scanpy.external as sce
from sklearn import linear_model
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay
from scipy.io import mmread
from scipy.sparse import csc_matrix, diags
from scipy.sparse.linalg import svds
import gc
import sys
from sklearn.neighbors import NearestNeighbors



def fit_knn(mat_train, mat_holdout, n_neighbors, algorithm = 'kd_tree'):
    
    # fit knn using mat_train
    # return nn indices and distances in train set for holdout set
    knn = NearestNeighbors(n_neighbors = n_neighbors, algorithm = algorithm).fit(mat_train)
    distances, indices = knn.kneighbors(mat_holdout)
    indices = indices[:,1:]
    distances = distances[:,1:]

    return indices, distances



def calc_knn_prop(knn_indices, labels_train, label_categories):

    # knn_indices: shape = (n_holdout_samples, (knn-1)), np.array
    # labels_train: shape = (n_train_samples, ), pd.object
    # label_categories: shape = (n_label_categories, ), np.array
    n = knn_indices.shape[0]
    n_category = label_categories.shape[0]
    nn_prop = np.zeros(shape = (n, n_category))

    for i in range(n):
        knn_labels = labels_train[knn_indices[i,]]
        for k in range(n_category):
            nn_prop[i, k] = sum(knn_labels == label_categories[k]) 

    nn_prop = nn_prop / knn_indices.shape[1]
    return nn_prop


def calc_oobNN(adata_orig, batch_key, condition_key, n_neighbors=15):
    ''' Compute out-of-batch k-nearest-neighbor composition 

    Parameters
    ----------
    adata_orig : anndata object
        Expression data stored in adata_orig.X, based on which to compute out of batch nearest neighbors.
    batch_key : str
        Variable name indicating batch. Should be a column name of adata.obs.
    condition_key : str
        Variable name indicating condition. Should be a column name of adata.obs. 
        We compute out-of-batch proportion of each condition level within each cell's neighborhood.
    n_neighbors : int, optional
        Number of k-nearest neighbors.
    
    Returns
    ----------
    res: anndata object
        One new attribute added. 
        res.obsm['knn_prop'] : pd.DataFrame, out-of-batch k-nearest-neighbor composition, cell-by-condition
    '''

    np.random.seed(123)
    
    list_holdout = []
    for holdout_idx in np.unique(adata_orig.obs[batch_key]):

        adata_train = adata_orig[~adata_orig.obs[batch_key].isin([holdout_idx])]
        adata_holdout = adata_orig[adata_orig.obs[batch_key].isin([holdout_idx])]
        num_cells = adata_train.obs[condition_key].value_counts().min()
    
        a_list = []
        for x in np.unique(adata_train.obs[condition_key]):
            a1 = adata_train[adata_train.obs[condition_key].isin([x])]
            random_indices = np.random.choice(a1.shape[0], size=num_cells, replace = False)
            a1 = a1[random_indices,:]
            a_list.append(a1)
 
        adata_train = ad.concat(a_list)
        adata = ad.concat([adata_train, adata_holdout])
    
        mat = sc.tl.pca(adata.X, n_comps = 20)
        mat_train = mat[~adata.obs[batch_key].isin([holdout_idx]),]
        mat_holdout = mat[adata.obs[batch_key].isin([holdout_idx]),]
    
        # fit knn
        indices, distances = fit_knn(mat_train=mat_train, mat_holdout=mat_holdout, n_neighbors=n_neighbors, algorithm = 'kd_tree')

        # compute proprotion
        labels_train = adata_train.obs[condition_key].astype('object')
        label_categories = np.unique(labels_train)
        result = calc_knn_prop(indices, labels_train, label_categories)
        knn_df = pd.DataFrame(data=result, 
                              index =  adata_holdout.obs_names,
                              columns = label_categories)
        adata_holdout.obsm['knn_prop'] = knn_df
        list_holdout.append(adata_holdout)

    res = ad.concat(list_holdout)
    return res





r"""
CyLib: High-performance computational library for omics data.

This module provides optimized computational utilities and fast implementations
of common algorithms used throughout the omicverse ecosystem. It includes both
Python and Cython-optimized functions for performance-critical operations.

Key features:
- Fast numerical computations for large-scale data
- Optimized algorithms for single-cell analysis
- Analytics and performance monitoring utilities
- Memory-efficient data structures
- GPU acceleration support where available

Main components:
    fast_utils: High-performance utility functions
    _analytics_sender: Analytics and telemetry utilities

Applications:
- Accelerated preprocessing of large datasets
- Fast neighborhood calculations
- Efficient similarity computations
- Performance monitoring and optimization
- Memory-efficient data transformations

Performance features:
- Cython-compiled critical functions
- Vectorized numpy operations
- Sparse matrix optimizations
- Memory mapping for large files
- Parallel processing support

Examples:
    >>> import omicverse as ov
    >>> # Fast utility functions are automatically available
    >>> # through other omicverse modules
    >>> 
    >>> # Direct usage (advanced)
    >>> from omicverse.cylib import fast_utils
    >>> result = fast_utils.fast_computation(data)

Notes:
    - Functions are primarily used internally by other modules
    - Provides significant speedups for large datasets
    - Maintains compatibility with standard numpy/scipy interfaces
    - Some functions may require compilation of Cython extensions
"""

from .fast_utils import *


r"""
OmicVerse: A comprehensive omic framework for multi-omic analysis.

OmicVerse is a Python package that provides a unified framework for analyzing 
multi-omic data including single-cell RNA-seq, bulk RNA-seq, spatial transcriptomics,
ATAC-seq, and multi-modal integration. It offers streamlined workflows for
preprocessing, analysis, visualization, and interpretation of complex biological data.

Main modules:
    bulk: Bulk RNA-seq analysis including differential expression, pathway analysis
    single: Single-cell RNA-seq analysis including clustering, trajectory inference
    bulk2single: Deconvolution and mapping between bulk and single-cell data
    space: Spatial transcriptomics analysis and integration
    pp: Preprocessing utilities for quality control and normalization
    pl: Comprehensive plotting and visualization functions
    utils: Utility functions for data handling and analysis
    popv: Population-level variation analysis tools

Key features:
    - Unified API for multiple omics data types
    - GPU acceleration support for large-scale analysis
    - Extensive visualization capabilities
    - Integration with popular bioinformatics tools
    - Comprehensive documentation and tutorials

Examples:
    >>> import omicverse as ov
    >>> adata = ov.read('data.h5ad')
    >>> ov.pp.preprocess(adata)
    >>> ov.single.leiden(adata)
    >>> ov.pl.umap(adata, color='leiden')
"""

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    from pkg_resources import get_distribution
    version = lambda name: get_distribution(name).version

# Core submodules
from . import bulk, single, utils, bulk2single, pp, space, pl

# Optional PopV module requires ``scvi-tools``. Skip if dependency missing.

#usually
from .utils._data import read
from .utils._plot import palette,ov_plot_set,plot_set


name = "omicverse"
try:
    __version__ = version(name)
except Exception:
    __version__ = "unknown"


from ._settings import settings,generate_reference_table


# 导入 matplotlib.pyplot
import matplotlib.pyplot as plt

# 将 plt 作为 omicverse 的一个属性
plt = plt  # 注意：确保没有其他变量名冲突

import numpy as np

np = np  # 注意：确保没有其他变量名冲突

# 导入 pandas
import pandas as pd

# 将 pd 作为 omicverse 的一个属性
pd = pd  # 注意：确保没有其他变量名冲突